#!/usr/bin/env python

""" MultiQC module to parse similarity matrix output by sourmash gather """

import csv
import logging

from multiqc.plots import bargraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


class GatherMixin:
    def parse_gather(self):
        """
        Modeled after the kraken module and the sourmash compare module.
        """

        self.top_n = getattr(config, "sourmash", {}).get("gather", {}).get("top_n", 5)

        # find and load gather reports
        self.gather_raw_data = dict()
        for f in self.find_log_files("sourmash/gather", filehandles=True):
            data = []
            for line in csv.DictReader(f["f"]):
                data.append(
                    {
                        "query_name": line["query_name"],
                        "match_name": line["name"],
                        "pct_unique_weighted": float(line["f_unique_weighted"]) * 100,
                        "pct_match": float(line["f_match"]) * 100,
                    }
                )
            if data:
                self.gather_raw_data[f["s_name"]] = data
            self.add_data_source(f, section="gather")

        self.gather_raw_data = self.ignore_samples(self.gather_raw_data)

        if len(self.gather_raw_data) == 0:
            return 0

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.gather_raw_data, "multiqc_sourmash_gather")

        # initialize variables to store summarized information
        self.gather_pct_per_match_all_samples = dict()
        self.gather_pct_unclassified_per_sample = dict()
        self.gather_pct_top_five_per_sample = dict()
        self.gather_top_five_matches = []

        # run functions to summarize information
        self.calculate_pct_per_match_all_samples()
        self.calculate_pct_unclassified_per_sample()
        self.calculate_pct_top_five_per_sample()

        # run functions to build multiqc report components
        self.general_stats_columns()
        self.top_five_barplot()

        # return the number of reports id'd, which will be logged in sourmash.py
        return len(self.gather_raw_data)

    def calculate_pct_per_match_all_samples(self):
        """
        Sum the percent of each genome match across queries,
        e.g. if "E. coli K12" is found to be 10% in sample A, and 20% in sample B, the
        output would be "E. coli K12": 30".

        The output is a dictionary with `match_names` (genomes) as keys, and summed
        percents as values. These values are used to identify the top N matches
        identified across all samples.
        """
        for s_name, data in self.gather_raw_data.items():
            for row in data:
                match_name = row["match_name"]
                # loop over all samples and sum different variables
                if match_name not in self.gather_pct_per_match_all_samples:
                    self.gather_pct_per_match_all_samples[match_name] = 0
                self.gather_pct_per_match_all_samples[match_name] += row["pct_unique_weighted"]

    def calculate_pct_unclassified_per_sample(self):
        """
        Calculate the percent of each sample that was unclassified.
        """
        for s_name, data in self.gather_raw_data.items():
            for row in data:
                # sum over all matches for a given query to calculate the percent classified
                if s_name not in self.gather_pct_unclassified_per_sample:
                    self.gather_pct_unclassified_per_sample[s_name] = 0
                self.gather_pct_unclassified_per_sample[s_name] += row["pct_unique_weighted"]
            # convert to % unclassified
            self.gather_pct_unclassified_per_sample[s_name] = 100 - self.gather_pct_unclassified_per_sample[s_name]

    def calculate_pct_top_five_per_sample(self):
        """
        Calculate the percent of each sample that is attributable to the top genomes
        across all samples. The output is a dictionary with samples as keys and
        the summed percent for the top genomes as values.
        """
        # get top genomes matched across samples
        sorted_pct = sorted(self.gather_pct_per_match_all_samples.items(), key=lambda x: x[1], reverse=True)
        for pct_sum in sorted_pct[: self.top_n]:
            self.gather_top_five_matches.append(pct_sum[0])  # append the genome name to the top five list

        # calculate the pct attributable to the top matches per sample
        for match_name in self.gather_top_five_matches:
            for s_name, data in self.gather_raw_data.items():
                if s_name not in self.gather_pct_top_five_per_sample:
                    self.gather_pct_top_five_per_sample[s_name] = 0
                for row in data:
                    if row["match_name"] == match_name:
                        self.gather_pct_top_five_per_sample[s_name] += row["pct_unique_weighted"]

    def general_stats_columns(self):
        """
        add columns to the general statistics table for % top matches and % unclassified
        """
        # Column headers
        headers = {
            f"% Top {self.top_n}": {
                "title": f"% Top {self.top_n} Genomes",
                "description": f"Percentage of sample that was classified as one of the top {self.top_n} genomes ({', '.join(self.gather_top_five_matches)})",
                "suffix": "%",
                "max": 100,
                "scale": "PuBu",
            },
            "% Unclassified": {
                "title": "% Unclassified",
                "description": "Percentage of sample that was unclassified",
                "suffix": "%",
                "max": 100,
                "scale": "OrRd",
            },
        }

        # set table data
        tdata = {}
        for s_name, data in self.gather_raw_data.items():
            tdata[s_name] = {}
            tdata[s_name]["% Unclassified"] = self.gather_pct_unclassified_per_sample[s_name]
            tdata[s_name][f"% Top {self.top_n}"] = self.gather_pct_top_five_per_sample[s_name]

        self.general_stats_addcols(tdata, headers)

    def top_five_barplot(self):
        """
        Add a bar plot showing the percentage of top genomes, the percentage of other
        genomes, and the unclassified percentage
        """

        pd = []  # plot data
        # A list of categories to be shown as a color on the plot. includes the
        # top genomes, other, and unclassified
        cats = list()
        pconfig = {
            "id": "sourmash-gather-top-plot",
            "title": "Sourmash gather: top genomes",
            "ylab": f"% of sample covered by top {self.top_n} genomes",
            # do not show the 'Counts / Percentages' switch, since gather only reports
            # percentages
            "cpswitch": False,
            # Initial display should show percentage, not counts
            "cpswitch_c_active": False,
        }

        # create dictionaries for plot fill names and data (match percentages)
        match_cats = {}
        match_data = {}

        pct_shown = {}
        for match_name in self.gather_top_five_matches:
            # make match_name a fill category for the plot
            match_cats[match_name] = {"name": match_name}
            # pull out percents for this match_name from each sample
            for s_name, data in self.gather_raw_data.items():
                if s_name not in match_data:
                    match_data[s_name] = dict()
                if s_name not in pct_shown:
                    pct_shown[s_name] = 0
                for row in data:
                    if row["match_name"] == match_name:
                        if match_name not in match_data[s_name]:
                            match_data[s_name][match_name] = 0
                        match_data[s_name][match_name] += row["pct_unique_weighted"]
                        pct_shown[s_name] += row["pct_unique_weighted"]

        # Add in unclassified pct and other accounted for by genomes matches not in the top
        for s_name, data in self.gather_raw_data.items():
            # unclassified
            match_data[s_name]["unclassified"] = self.gather_pct_unclassified_per_sample[s_name]
            pct_shown[s_name] += self.gather_pct_unclassified_per_sample[s_name]
            # other
            match_data[s_name]["other"] = 100 - (
                self.gather_pct_unclassified_per_sample[s_name] + self.gather_pct_top_five_per_sample[s_name]
            )

        match_cats["other"] = {"name": "Other", "color": "#cccccc"}
        match_cats["unclassified"] = {"name": "Unclassified", "color": "#d4949c"}

        cats.append(match_cats)
        pd.append(match_data)

        self.add_section(
            name="Gather: Top Genomes",
            anchor="sourmash-gather-top",
            description=f"The percentage of the sample falling into the top {self.top_n} genome matches.",
            helptext=f"""
                To make this plot, the percentage of each sample assigned to a given genome is summed across all samples.
                The percentages for these top {self.top_n} genomes are then plotted, as well as the unclassified percentage.
                The unclassified count is always shown across all taxa ranks.
                The category _"Other"_ shows the difference between the above total percentages and the sum of the percentages
                in the top {self.top_n} genomes shown + unclassified. This should cover all genomes _not_ in the top {self.top_n}, +/- any rounding errors.
            """,
            plot=bargraph.plot(pd, cats, pconfig),
        )



import logging
from multiqc import config
from multiqc.plots import bargraph

log = logging.getLogger(__name__)


def parse_reports(self):
    # To store the summary data
    self.stats = dict()

    # Parse the output files
    parse_stat_files(self)

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    self.add_software_version(None)

    # Remove filtered samples
    self.stats = self.ignore_samples(self.stats)

    if self.stats:
        self.write_data_file(self.stats, "multiqc_humid_stats")
        add_general_stats(self)
        add_stats_section(self)


def parse_stat_files(self):
    for f in self.find_log_files("humid/stats", filehandles=True):
        # There is no sample name in the log, so we use the root of the
        # file as sample name (since the filename is always stats.dat
        s_name = self.clean_s_name(f["root"], f)

        # Read the statistics from file
        d = {}
        for line in f["f"]:
            try:
                field, value = line.strip().split(": ")
                d[field] = int(value)
            except ValueError:
                pass

        try:
            # Calculate additional statistics
            d["filtered"] = d["total"] - d["usable"]
            d["duplicates"] = d["total"] - d["clusters"] - d["filtered"]

            # Make sure we only return data that makes sense
            if not sum(d[field] for field in ["duplicates", "clusters", "filtered"]) == d["total"]:
                log.warning(f"HUMID stats looked wrong, skipping: {s_name}")
                continue
        except KeyError as e:
            log.warning(f"Expected HUMID keys missing {e}, skipping sample: {s_name}")
            continue

        # Got this far, data must be good
        if s_name in self.stats:
            log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
        self.stats[s_name] = d
        self.add_data_source(f, s_name)


def add_general_stats(self):
    def add_unique_reads(self):
        """
        Add the number of reads after deduplication

        This corresponds to the number of clusters in HUMID
        """
        data = {k: {"uniq": v["clusters"]} for k, v in self.stats.items()}
        headers = dict()
        headers["uniq"] = {
            "title": f"{config.read_count_prefix} Unique Reads",
            "description": f"Reads remaining after deduplication ({config.read_count_desc})",
            "shared_key": "read_count",
            "modify": lambda x: x * config.read_count_multiplier,
            "scale": "Blues",
        }
        self.general_stats_addcols(data, headers)

    def add_passing_deduplication(self):
        """
        Add the percentage of reads remaining after deduplication
        """
        data = {k: {"perc": (v["clusters"] / v["total"]) * 100} for k, v in self.stats.items()}
        headers = dict()
        headers["perc"] = {
            "title": "% Pass Dedup",
            "description": "% processed reads that passed deduplication",
            "suffix": "%",
            "min": 0,
            "max": 100,
            "scale": "RdBu",
        }
        self.general_stats_addcols(data, headers)

    add_unique_reads(self)
    add_passing_deduplication(self)


def add_stats_section(self):
    # The values we want to plot (add to the total number of reads)
    cats = dict()
    cats["clusters"] = {"name": "Unique reads"}
    cats["duplicates"] = {"name": "Duplicate reads"}
    cats["filtered"] = {"name": "Filtered reads"}

    # Bargraph configuration
    config = {
        "id": "humid-bargraph",
        "title": "HUMID: Deduplication results",
        "ylab": "Number of reads",
        "hide_zero_cats": False,
    }
    self.add_section(
        name="Duplication Summary",
        anchor="humid-section",
        description="""
            Duplication statistics per sample. Every read in the
            input data has been assigned to one of the three categories
            shown here.
            """,
        helptext="""
            - **Unique reads** are the reads that are left over after deduplication.
            - **Duplicate reads** are the reads that were determined to be duplicates of the **Unique reads**.
            - **Filtered reads** were reads that could not be analysed, due to N nucleotides, or because the UMI or sequences were too short to use.
            """,
        plot=bargraph.plot(self.stats, cats, config),
    )



""" MultiQC module to parse output from FastQ Screen """


import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"Fastq_screen version: ([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="FastQ Screen",
            anchor="fastq_screen",
            href="http://www.bioinformatics.babraham.ac.uk/projects/fastq_screen/",
            info="allows you to screen a library of sequences in FastQ format against"
            " a set of sequence databases so you can see if the composition of the"
            " library matches with what you expect.",
            doi="10.12688/f1000research.15931.2",
        )

        # Find and load any FastQ Screen reports
        self.fq_screen_data = dict()
        self.num_orgs = 0
        for f in self.find_log_files("fastq_screen", filehandles=True):
            parsed_data = self.parse_fqscreen(f)
            if parsed_data is not None:
                if f["s_name"] in self.fq_screen_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
                self.add_data_source(f)
                self.fq_screen_data[f["s_name"]] = parsed_data

        # Filter to strip out ignored sample names
        self.fq_screen_data = self.ignore_samples(self.fq_screen_data)

        if len(self.fq_screen_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.fq_screen_data)} reports")

        # Section 1 - Alignment Profiles
        self.add_section(name="Mapped Reads", anchor="fastq_screen_mapped_reads", plot=self.fqscreen_simple_plot())

        # Section 2 - Optional bisfulfite plot
        self.fqscreen_bisulfite_plot()

        # Write the total counts and percentages to files
        self.write_data_file(self.parse_csv(), "multiqc_fastq_screen")

    def parse_fqscreen(self, f):
        """Parse the FastQ Screen output into a 3D dict"""
        parsed_data = {}
        nohits_pct = None
        headers = None
        bs_headers = None
        for line in f["f"]:
            # Skip comment lines
            if line.startswith("#"):
                version_match = re.search(VERSION_REGEX, line)
                if version_match:
                    self.add_software_version(version_match.group(1), f["s_name"])
                continue
            if line.startswith("%Hit_no_genomes:") or line.startswith("%Hit_no_libraries:"):
                nohits_pct = float(line.split(":", 1)[1])
                parsed_data["No hits"] = {"percentages": {"one_hit_one_library": nohits_pct}}
            else:
                s = line.strip().split("\t")

                # Regular FastQ Screen table section
                if len(s) == 12:
                    if headers is None:
                        headers = s
                    else:
                        # Can't use #Reads in subset as varies. #Reads_processed should be same for all orgs in a sample
                        parsed_data["total_reads"] = int(s[1])
                        # Loop through all columns
                        parsed_data[s[0]] = {"percentages": {}, "counts": {}}
                        for idx, h in enumerate(headers):
                            if idx == 0:
                                continue
                            dtype = "percentages" if h.startswith("%") else "counts"
                            field = (
                                h.replace("%", "")
                                .replace("#", "")
                                .replace("genomes", "libraries")
                                .replace("genome", "library")
                                .lower()
                            )
                            parsed_data[s[0]][dtype][field] = float(s[idx])

                # Optional Bisulfite table section
                elif len(s) == 9:
                    if bs_headers is None:
                        bs_headers = s
                    else:
                        # Loop through all columns
                        parsed_data[s[0]]["bisulfite_percentages"] = {}
                        parsed_data[s[0]]["bisulfite_counts"] = {}
                        for idx, h in enumerate(bs_headers):
                            if idx == 0:
                                continue
                            dtype = "bisulfite_percentages" if h.startswith("%") else "bisulfite_counts"
                            field = h.replace("%", "").replace("#", "").lower()
                            parsed_data[s[0]][dtype][field] = float(s[idx])

        if len(parsed_data) == 0:
            return None

        # Calculate no hits counts
        if parsed_data.get("total_reads") is not None and nohits_pct is not None:
            parsed_data["No hits"]["counts"] = {
                "one_hit_one_library": int((nohits_pct / 100.0) * float(parsed_data["total_reads"]))
            }
        else:
            log.warning(f"Couldn't find number of reads with no hits for '{f['s_name']}'")

        self.num_orgs = max(len(parsed_data), self.num_orgs)
        return parsed_data

    def parse_csv(self):
        totals = dict()
        for s in sorted(self.fq_screen_data.keys()):
            totals[s] = dict()
            for org in self.fq_screen_data[s]:
                if org == "total_reads":
                    totals[s]["total_reads"] = self.fq_screen_data[s][org]
                    continue
                try:
                    k = f"{org} counts"
                    totals[s][k] = self.fq_screen_data[s][org]["counts"]["one_hit_one_library"]
                    totals[s][k] += self.fq_screen_data[s][org]["counts"].get("multiple_hits_one_library", 0)
                    totals[s][k] += self.fq_screen_data[s][org]["counts"].get("one_hit_multiple_libraries", 0)
                    totals[s][k] += self.fq_screen_data[s][org]["counts"].get("multiple_hits_multiple_libraries", 0)
                except KeyError:
                    pass
                try:
                    k = f"{org} percentage"
                    totals[s][k] = self.fq_screen_data[s][org]["percentages"]["one_hit_one_library"]
                    totals[s][k] += self.fq_screen_data[s][org]["percentages"].get("multiple_hits_one_library", 0)
                    totals[s][k] += self.fq_screen_data[s][org]["percentages"].get("one_hit_multiple_libraries", 0)
                    totals[s][k] += self.fq_screen_data[s][org]["percentages"].get(
                        "multiple_hits_multiple_libraries", 0
                    )
                except KeyError:
                    pass
        return totals

    def fqscreen_simple_plot(self):
        """Makes a simple bar plot with summed alignment counts for
        each species, stacked."""

        # First, sum the different types of alignment counts
        data = {}
        org_counts = {}
        for s_name in sorted(self.fq_screen_data):
            data[s_name] = dict()
            sum_alignments = 0
            for org in self.fq_screen_data[s_name]:
                if org == "total_reads":
                    continue
                try:
                    data[s_name][org] = self.fq_screen_data[s_name][org]["counts"]["one_hit_one_library"]
                except KeyError:
                    log.error(
                        "No counts found for '{}' ('{}'). Could be malformed or very old FastQ Screen results. Skipping sample".format(
                            org, s_name
                        )
                    )
                    continue
                try:
                    data[s_name][org] += self.fq_screen_data[s_name][org]["counts"]["multiple_hits_one_library"]
                except KeyError:
                    pass
                sum_alignments += data[s_name][org]
                org_counts[org] = org_counts.get(org, 0) + data[s_name][org]

            # Calculate hits in multiple genomes
            if "total_reads" in self.fq_screen_data[s_name]:
                data[s_name]["Multiple Genomes"] = self.fq_screen_data[s_name]["total_reads"] - sum_alignments

        # Sort the categories by the total read counts
        cats = dict()
        for org in sorted(org_counts, key=org_counts.get, reverse=True):
            if org not in cats and org != "No hits":
                cats[org] = {"name": org}

        # Strip empty dicts
        for s_name in list(data.keys()):
            if len(data[s_name]) == 0:
                del data[s_name]

        pconfig = {
            "id": "fastq_screen_plot",
            "title": "FastQ Screen: Mapped Reads",
            "cpswitch_c_active": False,
            "ylab": "Mapped reads",
        }
        cats["Multiple Genomes"] = {"name": "Multiple Genomes", "color": "#820000"}
        cats["No hits"] = {"name": "No hits", "color": "#cccccc"}

        return bargraph.plot(data, cats, pconfig)

    def fqscreen_bisulfite_plot(self):
        """Make a stacked barplot for the bisulfite data, if we have any"""

        pconfig = {
            "id": "fastq_screen_bisulfite_plot",
            "title": "FastQ Screen: Bisulfite Mapping Strand Orientation",
            "hide_zero_cats": False,
            "ylab": "Percentages",
            "data_labels": [],
        }

        cats = {
            "original_top_strand": {"name": "Original top strand", "color": "#80cdc1"},
            "complementary_to_original_top_strand": {
                "name": "Complementary to original top strand",
                "color": "#018571",
            },
            "complementary_to_original_bottom_strand": {
                "name": "Complementary to original bottom strand",
                "color": "#a6611a",
            },
            "original_bottom_strand": {"name": "Original bottom strand", "color": "#dfc27d"},
        }

        # Pull out the data that we want
        pdata_unsorted = {}
        org_counts = {}
        max_count = 0
        for s_name, d in self.fq_screen_data.items():
            for org, dd in d.items():
                if org not in pdata_unsorted:
                    pdata_unsorted[org] = {}
                    org_counts[org] = 0
                try:
                    pdata_unsorted[org][s_name] = dd["bisulfite_counts"]
                    total_counts = sum([c for c in dd["bisulfite_counts"].values()])
                    org_counts[org] += total_counts
                    max_count = max(max_count, total_counts)
                except (TypeError, KeyError):
                    pass

        # Consistent y-max across all genomes
        pconfig["ymax"] = max_count

        # Show tabbed bar plots, sorted by total read count
        pdata = []
        pcats = []
        for org in sorted(org_counts, key=org_counts.get, reverse=True):
            if org_counts[org] > 0:
                pconfig["data_labels"].append(org)
                pdata.append(pdata_unsorted[org])
                pcats.append(cats)

        if len(pdata) == 0:
            return None

        self.add_section(
            name="Bisulfite Reads", anchor="fastq_screen_bisulfite", plot=bargraph.plot(pdata, pcats, pconfig)
        )



""" MultiQC module to parse output from MALT """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

log = logging.getLogger(__name__)

VERSION_REGEX = r"Version\s+MALT \(version ([\d\.]+),.*"


class MultiqcModule(BaseMultiqcModule):
    """Malt Module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="MALT",
            anchor="malt",
            href="http://ab.inf.uni-tuebingen.de/software/malt/",
            info="performs alignment of metagenomic reads against a database of reference sequences (such as NR, GenBank or Silva) and produces a MEGAN RMA file as output.",
            doi="10.1101/050559 ",
        )

        # Find and load Malt reports
        self.malt_data = dict()

        for f in self.find_log_files("malt", filehandles=True):
            self.parse_logs(f)

        self.malt_data = self.ignore_samples(self.malt_data)

        if len(self.malt_data) == 0:
            raise ModuleNoSamplesFound

        # Write data to file
        self.write_data_file(self.malt_data, "malt")

        log.info(f"Found {len(self.malt_data)} reports")

        self.malt_general_stats()
        self.mappability_barplot()
        self.taxonomic_assignation_barplot()

    def parse_logs(self, f):
        """Parses a Malt log file"""
        reading = False
        keys = [
            "Total reads",
            "Assig. Taxonomy",
            "Num. of queries",
            "Aligned queries",
            "Num. alignments",
        ]
        version = None
        for line in f["f"]:
            line = line.rstrip()
            if line.startswith("Version"):
                version_match = re.search(VERSION_REGEX, line)
                if version_match:
                    version = version_match.group(1)

            if line.startswith("+++++ Aligning file:") and reading is False:
                reading = True
                s_name = line.split()[-1]
                s_name = self.clean_s_name(s_name, f)
                if s_name in self.malt_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.add_data_source(f, s_name=s_name)
                self.malt_data[s_name] = {}
                if version is not None:
                    self.add_software_version(version, s_name)

            elif reading:
                for k in keys:
                    if line.startswith(k):
                        self.malt_data[s_name][k] = int(line.split()[-1].replace(",", ""))
                        if k == "Num. alignments":
                            try:
                                self.malt_data[s_name]["Non mapped"] = (
                                    self.malt_data[s_name]["Num. of queries"] - self.malt_data[s_name]["Total reads"]
                                )
                                self.malt_data[s_name]["No Assig. Taxonomy"] = (
                                    self.malt_data[s_name]["Total reads"] - self.malt_data[s_name]["Assig. Taxonomy"]
                                )
                                try:
                                    self.malt_data[s_name]["Mappability"] = (
                                        float(self.malt_data[s_name]["Total reads"])
                                        / float(self.malt_data[s_name]["Num. of queries"])
                                    ) * 100.0
                                except ZeroDivisionError:
                                    self.malt_data[s_name]["Mappability"] = 0
                                try:
                                    self.malt_data[s_name]["Taxonomic assignment success"] = (
                                        float(self.malt_data[s_name]["Assig. Taxonomy"])
                                        / float(self.malt_data[s_name]["Total reads"])
                                    ) * 100.0
                                except ZeroDivisionError:
                                    self.malt_data[s_name]["Taxonomic assignment success"] = 0
                            except KeyError:
                                pass
                            reading = False

    def mappability_barplot(self):
        """Mappability barplot

        Mappability = (Total reads / Num. of queries) * 100
        """
        cats = {
            "Total reads": {"name": "Mapped reads"},
            "Non mapped": {"name": "Non Mapped reads"},
        }
        config = {
            "id": "malt-mappability-plot",
            "title": "MALT: Metagenomic Mappability",
            "ylab": "Read Counts",
        }
        self.add_section(
            name="Metagenomic Mappability",
            anchor="malt-mappability",
            description="Number of mapped reads.",
            plot=bargraph.plot(self.malt_data, cats, config),
        )

    def taxonomic_assignation_barplot(self):
        """Taxonomic assignment barplot

        Taxonomic assignment success = (Assig. Taxonomy / Total reads) * 100
        """
        cats = ["Assig. Taxonomy", "No Assig. Taxonomy"]
        config = {
            "id": "malt-taxonomic-success-plot",
            "title": "MALT: Taxonomic assignment success",
            "ylab": "Read Counts",
        }
        self.add_section(
            name="Taxonomic assignment success",
            anchor="malt-taxonomic-success",
            description="Shows the number of mapped reads assigned to a taxonomic node.",
            plot=bargraph.plot(self.malt_data, cats, config),
        )

    def malt_general_stats(self):
        """MALT General Statistics table"""
        headers = {
            "Taxonomic assignment success": {
                "title": "% Tax assigned",
                "description": "Percentage of mapped reads assigned to a taxonomic node",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
            "Assig. Taxonomy": {
                "title": f"{config.read_count_prefix} Tax assigned",
                "description": f"Number of reads assigned to a Taxonomic node ({config.read_count_desc})",
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "hidden": True,
            },
            "Mappability": {
                "title": "% Metagenomic Mapped",
                "description": "Percentage of mapped reads",
                "suffix": "%",
                "max": 100,
                "scale": "RdYlGn",
            },
            "Total reads": {
                "title": f"{config.read_count_prefix} Mapped",
                "description": f"Number of mapped reads ({config.read_count_desc})",
                "scale": "PuBu",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "Num. of queries": {
                "title": f"{config.read_count_prefix} Reads",
                "description": f"Number of reads in sample ({config.read_count_desc})",
                "scale": "Purples",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
        }
        self.general_stats_addcols(self.malt_data, headers)



""" MultiQC module to parse output from gopeaks """

import json
import logging
from pathlib import Path

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """gopeaks module"""

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="GoPeaks",
            anchor="gopeaks",
            href="https://github.com/maxsonBraunLab/gopeaks",
            info="is designed to call peaks from aligned CUT&TAG sequencing reads.\
                  Gopeaks uses a binomial distribution to model the read counts \
                  in sliding windows across the genome and calculate peak regions \
                  that are enriched over the background.",
            doi="10.1186/s13059-022-02707-w",
        )

        # data vars ---------------------------------------------------------------------

        # Find and load any gopeaks reports
        self.gopeaks_data = dict()
        for f in self.find_log_files("gopeaks", filehandles=True):
            parsed = self.parse_gopeaks_log(f)

            if parsed is not None:
                if f["s_name"] in self.gopeaks_data:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {f['s_name']}")

                if "gopeaks_version" in parsed:
                    self.add_software_version(parsed["gopeaks_version"], f["s_name"])

                self.gopeaks_data[f["s_name"]] = parsed

                # add gopeaks data to multiqc_source.txt
                self.add_data_source(f)

        # filter away samples if MultiQC user does not want them
        self.gopeaks_data = self.ignore_samples(self.gopeaks_data)

        if len(self.gopeaks_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.gopeaks_data)} samples")

        self.write_data_file(self.gopeaks_data, "multiqc_gopeaks")

        # Add sample log info to basic stats table
        self.gopeaks_general_stats_table()

        # Add sample log info to bargraph
        self.gopeaks_bargraph()

    # parsing functions -------------------------------------------------------------

    def parse_gopeaks_log(self, f):
        """
        Read gopeaks json log file and extract number of peaks.
        """
        with open(Path(f["root"]) / Path(f["fn"])) as f:
            sample_log = json.load(f)

        return sample_log

    def gopeaks_general_stats_table(self):
        """
        Put peak counts to the general table.
        """

        headers = {
            "peak_counts": {
                "title": "Peak Counts",
                "description": "Number of peaks per sample",
                "min": 0,
                "scale": "YlGnBu",
                "format": "{:,.0f}",
            }
        }
        self.general_stats_addcols(self.gopeaks_data, headers)

    def gopeaks_bargraph(self):
        """
        Put peak counts to a bargraph.
        """

        cats = {"peak_counts": {"name": "Peak Counts"}}
        config = {
            "id": "GoPeaksBarGraph",
            "title": "GoPeaks: Number of Peaks by Sample",
            "ylab": "Sample",
            "logswitch": True,
            "cpswitch": False,
        }

        self.add_section(
            name="GoPeaks",
            anchor="gopeaks_bargraph",
            description="Number of peaks called by GoPeaks.",
            plot=bargraph.plot(self.gopeaks_data, cats, config),
        )



""" MultiQC submodule to parse output from Picard CollectIlluminaLaneMetrics """

import logging
from collections import defaultdict

from multiqc.modules.picard import util
from multiqc.plots import table

# Initialise the logger
log = logging.getLogger(__name__)


def lane_metrics_table(module, data):
    headers = {
        "CLUSTER_DENSITY": {
            "title": "Cluster Density",
            "description": "The number of clusters per unit area on this lane (cluster / " "mm^2`)",
            "scale": "Greens",
        },
        "TYPE_NAME": {
            "title": "Read Number",
            "description": "Defines an Illumina template read number (first or second)",
            "modify": lambda x: x.lower(),
        },
        "PREPHASING_APPLIED": {
            "title": "Prephasing Applied",
            "description": "Median pre-phasing value across all tiles in a lane, applied "
            "to the first and second template reads",
            "scale": "BuPu",
            "max": 1,
        },
        "PHASING_APPLIED": {
            "title": "Phasing Applied",
            "description": "Median phasing value across all tiles in a lane, applied to "
            "the first and second template reads",
            "scale": "BuPu",
            "max": 1,
        },
    }

    table_config = {
        "id": f"{module.anchor}-illumina-lane-metrics-table",
        "namespace": module.name,
        "table_title": f"{module.name} Illumina Lane Metrics",
    }
    tdata = {}
    for run_name, run in data.items():
        for lane_number, lane in run.items():
            tdata[f"{run_name} - L{lane_number}"] = lane
    return table.plot(tdata, headers, table_config)


def parse_reports(module):
    """Find Picard IlluminaLaneMetrics reports and parse their data"""

    # There can be two types of these files for the same sample: one with IlluminaLaneMetrics,
    # and one with IlluminaPhasingMetrics. We want to collect both, thus using default dicts
    # and calling .update() on them when any metrics are found.
    data_by_lane_by_run = defaultdict(lambda: defaultdict(dict))

    # Go through logs and find Metrics
    for f in module.find_log_files("picard/collectilluminalanemetrics", filehandles=True):
        # Sample name from input file name by default
        run_name = f["s_name"]
        keys = None

        for line in f["f"]:
            maybe_run_name = util.extract_sample_name(
                module,
                line,
                f,
                picard_tool="CollectIlluminaLaneMetrics",
                picard_opt="OUTPUT_PREFIX",
            )
            if maybe_run_name:
                run_name = maybe_run_name
                keys = None

            if run_name is None:
                continue

            if util.is_line_right_before_table(line, picard_class=["IlluminaLaneMetrics", "IlluminaPhasingMetrics"]):
                keys = f["f"].readline().strip("\n").split("\t")
                if run_name in data_by_lane_by_run:
                    log.debug(f"Duplicate sample name found in {f['fn']}! Overwriting: {run_name}")
                module.add_data_source(f, s_name=run_name, section="IlluminaLaneMetrics")

            elif keys:
                vals = line.strip("\n").split("\t")
                if len(vals) != len(keys):
                    keys = None
                    continue

                d = dict(zip(keys, vals))
                lane = d["LANE"]
                data_by_lane_by_run[run_name][lane].update(d)

    data_by_lane_by_run = module.ignore_samples(data_by_lane_by_run)
    if len(data_by_lane_by_run) == 0:
        return 0

    # Superfluous function call to confirm that it is used in this module
    # Replace None with actual version if it is available
    module.add_software_version(None)

    # Write parsed data to a file
    module.write_data_file(data_by_lane_by_run, f"multiqc_{module.anchor}_IlluminaLaneMetrics")

    module.add_section(
        name="Lane Metrics",
        anchor="picard-illuminalanemetrics",
        description="Quality control metrics on cluster density for each lane of an "
        "Illumina flowcell. For more information, see the [Picard "
        "Documentation]("
        "https://broadinstitute.github.io/picard/picard-metric"
        "-definitions.html#IlluminaLaneMetrics).",
        plot=lane_metrics_table(module, data_by_lane_by_run),
    )

    # Return the number of detected samples to the parent module
    return len(data_by_lane_by_run)



""" MultiQC module to parse output from CCS """

import json
import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialse the parent object
        super(MultiqcModule, self).__init__(
            name="CCS",
            anchor="ccs",
            href="https://github.com/PacificBiosciences/ccs",
            info=" is used to generate highly accurate single-molecule consensus reads from PacBio sequencing.",
            # Can't find a DOI // doi=
        )

        # To store the mod data
        self.ccs_data = dict()
        self.parse_v4_log_files()
        self.parse_v5_log_files()
        self.ccs_data = self.ignore_samples(self.ccs_data)

        # If we found no data
        if not self.ccs_data:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.ccs_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        self.write_data_file(self.ccs_data, "multiqc_ccs_report")
        self.add_general_stats()
        self.add_sections()

    def parse_v4_log_files(self):
        for f in self.find_log_files("ccs/v4", filehandles=True):
            data = parse_PacBio_log(f["f"])
            v5_data = self.convert_to_v5(data)
            self.ccs_data[f["s_name"]] = v5_data
            self.add_data_source(f)

    def parse_v5_log_files(self):
        for f in self.find_log_files("ccs/v5", filehandles=True):
            v5_data = json.load(f["f"])
            self.ccs_data[f["s_name"]] = v5_data
            self.add_data_source(f)

    def add_general_stats(self):
        gstats_data = {}
        for s_name, attrs in self.ccs_data.items():
            gstats_data[s_name] = {}
            for attr in attrs["attributes"]:
                if "zmw_input" in attr["id"]:
                    gstats_data[s_name]["zmw_input"] = attr["value"]
                if "zmw_passed_yield" in attr["id"]:
                    gstats_data[s_name]["zmw_passed_yield"] = attr["value"]
            try:
                gstats_data[s_name]["zmw_pct_passed_yield"] = (
                    gstats_data[s_name]["zmw_passed_yield"] / gstats_data[s_name]["zmw_input"]
                ) * 100.0
            except (KeyError, ZeroDivisionError):
                pass

        headers = {
            "zmw_pct_passed_yield": {
                "title": "ZMWs %PF",
                "description": "Percent of ZMWs passing filters",
                "max": 100,
                "min": 0,
                "scale": "RdYlGn",
                "suffix": "%",
            },
            "zmw_passed_yield": {
                "title": "ZMWs PF",
                "description": "ZMWs pass filters",
                "scale": "BuGn",
                "format": "{.0f}",
                "shared_key": "zmw_count",
                "hidden": True,
            },
            "zmw_input": {
                "title": "ZMWs input",
                "description": "ZMWs input",
                "scale": "Purples",
                "format": "{.0f}",
                "shared_key": "zmw_count",
            },
        }
        self.general_stats_addcols(gstats_data, headers)

    def add_sections(self):
        # First we gather all the filters we encountered
        all_filters = dict()
        for s_name in self.ccs_data:
            for filter_reason in self.filter_and_pass(self.ccs_data[s_name]):
                all_filters[filter_reason] = 0

        # Then we add the counts for each filter to the plot data
        plot_data = dict()
        for s_name, data in self.ccs_data.items():
            plot_data[s_name] = dict()
            for reason in all_filters:
                for attribute in data["attributes"]:
                    if attribute["name"] == reason:
                        plot_data[s_name][reason] = attribute["value"]
                        all_filters[reason] += attribute["value"]
                        break
                # If we didn't find it, set to zero
                else:
                    plot_data[s_name][reason] = 0

        # Sort the categories based on the total counts
        # Smallest first, for Log10 view
        plot_cats = []
        for reason, count in sorted(all_filters.items(), key=lambda item: item[1], reverse=True):
            plot_cats.append(reason)

        # Plot configuration
        config = {
            "id": "ccs-filter-graph",
            "title": "CCS: ZMW results",
            "ylab": "Number of ZMWs",
            "xlab": "CCS report file",
            "logswitch": True,
        }

        self.add_section(
            name="ZMWs filtered and passed",
            anchor="ccs-filter",
            description="The number of ZMWs that failed or passed all filters for each of the subreads files.",
            helptext="""
                The number of ZMWs that passed all filters is shown as **ZMWs pass filters**. All other categories that
                are shown in the graph represent the number of ZMWs that were dropped for the specified reason.
            """,
            plot=bargraph.plot(plot_data, plot_cats, config),
        )

    def filter_and_pass(self, data):
        """Gather the reasons why ZMWs were failed or passed"""
        reasons = dict()

        # We only have to use the attributes
        attributes = data["attributes"]

        # Add filtere reasons (id starts with filtered_) and total passed
        for entry in attributes:
            if (
                entry["id"].startswith("filtered")
                or entry["id"] == "zmw_passed_yield"
                or entry["id"].startswith("ccs_processing.filtered")
                or entry["id"] == "ccs_processing.zmw_passed_yield"
            ):
                reasons[entry["name"]] = entry["value"]

        return reasons

    def convert_to_v5(self, data):
        """Convert the v4 format to the new CCS v5 json format"""
        # Initialise the v5 format dictionary
        v5 = dict()
        v5["id"] = "ccs_processing"
        attributes = list()
        v5["attributes"] = attributes

        # Update names for top level entries, they have been changed in v5
        data["ZMWs pass filters"] = data["ZMWs generating CCS"]
        data["ZMWs fail filters"] = data["ZMWs filtered"]

        # Add the top level values to the attributes
        for name in ("ZMWs input", "ZMWs pass filters", "ZMWs fail filters"):
            count = data[name]["count"]
            attributes.append(self.make_v5_entry(name, count))

        # Add the reasons for filtering to the attributes
        for reason in data["ZMWs filtered"]["Exclusive ZMW counts"]:
            count = data["ZMWs filtered"]["Exclusive ZMW counts"][reason]["count"]
            attributes.append(self.make_v5_entry(reason, count))

        return v5

    def make_v5_entry(self, name, count):
        """Make a v5 output entry based on name and count"""
        # Dictionary to map the report names to the v5 id annotation
        name_to_id = {
            "ZMWs input": "zmw_input",
            "ZMWs pass filters": "zmw_passed_yield",
            "ZMWs fail filters": "zmw_filtered_yield",
            "Below SNR threshold": "filtered_poor_snr",
            "Median length filter": "filtered_median_length_filter",
            "Lacking full passes": "filtered_too_few_passes",
            "Heteroduplex insertions": "filtered_heteroduplexes",
            "Coverage drops": "filtered_coverage_drops",
            "Insufficient draft cov": "filtered_insufficient_draft_cov",
            "Draft too different": "filtered_draft_too_different",
            "Draft generation error": "filtered_draft_failure",
            "Draft above --max-length": "filtered_draft_too_long",
            "Draft below --min-length": "filtered_draft_too_short",
            "Reads failed polishing": "filtered_read_failed_polish",
            "Empty coverage windows": "filtered_empty_window_during_polishing",
            "CCS did not converge": "filtered_non_convergent",
            "CCS below minimum RQ": "filtered_below_rq",
            "Unknown error": "filtered_unknown_error",
        }

        return {"name": name, "id": name_to_id[name], "value": count}


def parse_PacBio_log(file_content):
    """Parse ccs log file"""
    data = dict()
    # This is a local dictionary to store which annotations belong to which
    # result dictionary. This will be used to structure the output, but will
    # not be part of the output itself
    annotations = dict()
    current_annotation = None

    for line in file_content:
        # Get rid of trailing newlines
        line = line.strip("\n")
        # Did we enter a new section with annotations for an earlier result?
        # If so, we will only add an empty dictionary with the correct name
        # These field are of the format "something something for (A):"
        section_header_pattern = " for [(][A-Z][)][:]$"
        if re.search(section_header_pattern, line):
            linedata = parse_line(line)
            ann = linedata["annotation"]
            # Cut off the ' for (B):' part
            name = line[:-9]
            # We make a new heading with the current name under the data that
            # matches the current annotation
            current_annotation = dict()
            # We keep the dictonary accessible under 'current_annotation',
            # so we can keep adding new data to it without having to keep track
            # of where it belongs
            annotations[ann][name] = current_annotation
            continue

        linedata = parse_line(line)

        # If we got no data, we reached the end of the annotated section
        if not linedata:
            current_annotation = None
            continue

        # Lets get the name of the data
        name = linedata.pop("name")
        # If we are in an annotated section, we add the data to the current
        # annotation
        if current_annotation is not None:
            current_annotation[name] = linedata
        # Otherwise, we add the newfound annotation to the dictionary in case
        # we find a corresponding section later on.
        # The data from the current line we add directly to the output data
        else:
            annotation = linedata.pop("annotation")
            annotations[annotation] = linedata
            data[name] = linedata

    return data


def parse_line(line):
    """Parse a line from the ccs log file"""
    data = dict()

    # If we got an empty line to parse
    if not line.strip():
        return data

    # Split the line on the colon character
    key, values = line.strip().split(":")

    # The key can have multiple parts
    keys = key.strip().split()

    # Does the key have an annotation (A), (B) etc
    if re.fullmatch("[(][A-Z][)]", keys[-1]):
        # We store the annotation without the bracets
        data["annotation"] = keys[-1][1:-1]
        # And we add the rest of the key as name
        data["name"] = " ".join(keys[:-1])
    # Otherwise, we just store the name
    else:
        data["name"] = " ".join(keys)

    # Parsing the values
    values = values.strip().split()
    # Are there are no values we are done
    if not values:
        return data

    # If there is a single value
    if len(values) == 1:
        value = values.pop()
        # Is the value a percentage
        if value.endswith("%"):
            data["percentage"] = float(value[:-1])
        # Otherwise, it is an integer
        else:
            data["count"] = int(value)
    elif len(values) == 2:
        # If there are multiple values, they are in the format
        # count (percentage%)
        count = values[0]
        percentage = values[1]

        # The percentage should be in the format:
        # (12.34%) or (100%) or (0%) or (-nan%)
        # So we can remove the bracets first
        percentage = percentage[1:-1]
        # Then we make sure it is one of the supported patterns
        assert re.fullmatch(r"(\d+\.\d+%|\d+%|-nan%)", percentage)

        # If the percentage is this weird nan, set it to 0
        if percentage == "-nan%":
            data["percentage"] = 0.0
        else:  # Otherwise, we cut of the % sign and convert to float
            data["percentage"] = float(percentage[:-1])
        # Add the count as an integer
        data["count"] = int(count)

    return data



""" MultiQC module to parse output from DIAMOND """

import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"diamond v([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="DIAMOND",
            anchor="diamond",
            href="https://github.com/bbuchfink/diamond",
            info="a sequence aligner for protein and translated DNA searches, designed for high performance analysis of big sequence data.",
            doi="10.1038/s41592-021-01101-x",
        )

        # Find and load any DIAMOND reports
        self.diamond_data = dict()

        for f in self.find_log_files("diamond", filehandles=True):
            self.parse_logs(f)

        # Filter to strip out ignored sample names
        self.diamond_data = self.ignore_samples(self.diamond_data)

        if len(self.diamond_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.diamond_data)} reports")

        # Write parsed report data to file
        self.write_data_file(self.diamond_data, "diamond")
        self.diamond_general_stats()

    def parse_logs(self, f):
        """Parsing logs"""
        s_name = self.clean_s_name(f["root"], f)
        for line in f["f"]:
            # Try to get the sample name from --out, otherwise --query, fallback to directory name
            if "diamond blastx" in line and "--out" in line:
                s_name = line.split("--out ")[1].split(" ")[0]
                s_name = self.clean_s_name(s_name, f)
            elif "diamond blastx" in line and "--query" in line:
                s_name = line.split("--query ")[1].split(" ")[0]
                s_name = self.clean_s_name(s_name, f)

            # Get version
            version_match = re.search(VERSION_REGEX, line)
            if version_match:
                self.add_software_version(version_match.group(1), s_name)

            if "queries aligned" in line:
                self.add_data_source(f)
                if s_name in self.diamond_data:
                    log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                self.diamond_data[s_name] = {"queries_aligned": int(line.split(" ")[0])}

    def diamond_general_stats(self):
        """Diamond General Stats Table"""
        headers = {
            "queries_aligned": {
                "title": "Queries aligned",
                "description": "number of queries aligned",
                "scale": "YlGn",
                "format": "{:,.0f}",
            }
        }
        self.general_stats_addcols(self.diamond_data, headers)



""" MultiQC submodule to parse output from MinIONQC summary stats """

import copy
import logging
import os
import re

import yaml

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="MinIONQC",
            anchor="minionqc",
            href="https://github.com/roblanf/minion_qc",
            info=" is a QC tool for Oxford Nanopore sequencing data",
            doi="10.1093/bioinformatics/bty654",
        )

        # Find and load any minionqc reports
        self.minionqc_raw_data = dict()  # main dataset in original YAML format
        self.minionqc_data = dict()  # main dataset. Stats from all reads
        self.qfilt_data = dict()  # Stats from quality filtered reads
        self.q_threshold_list = set()  # quality thresholds
        for f in self.find_log_files("minionqc", filehandles=True):
            # get sample name
            s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=os.path.dirname(f["root"]))

            # parses minionqc summary data
            self.parse_minionqc_report(s_name, f)

        # Filter to strip out ignored sample names
        self.minionqc_data = self.ignore_samples(self.minionqc_data)
        if len(self.minionqc_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.minionqc_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # columns to present in MultiQC summary table
        headers = self.headers_to_use()

        # Just the total number of reads in the main General Statistics table
        self.general_stats_addcols(self.minionqc_data, {"total.reads": headers["total.reads"]})

        # writes parsed data into a file (by MultiQC)
        self.write_data_file(self.minionqc_data, "multiqc_minionqc")

        # tables and plots in order
        self.table_qALL()
        self.table_qfiltered()
        self.plot_readlengths()

    def parse_minionqc_report(self, s_name, f):
        """
        Parses minionqc's 'summary.yaml' report file for results.
        Uses only the "All reads" stats. Ignores "Q>=x" part.
        """
        try:
            summary_dict = yaml.safe_load(f["f"])
        except Exception as e:
            log.error(f"Error parsing MinIONQC input file {f['f']}: {e}")
            return

        # adds files used in MultiQC report
        self.add_data_source(f, s_name)

        # Do a deep copy as dicts are immutable
        self.minionqc_raw_data[s_name] = copy.deepcopy(summary_dict)

        # get q value threshold used for reads
        q_threshold = None
        for k in summary_dict.keys():
            if k.startswith("Q>="):
                q_threshold = k

        data_dict = {"all": summary_dict["All reads"], "q_filt": summary_dict[q_threshold]}

        for q_key in ["all", "q_filt"]:
            for key_1 in ["reads", "gigabases"]:
                for key_2 in data_dict[q_key][key_1]:
                    new_key = f"{key_1} {key_2}"
                    data_dict[q_key][new_key] = data_dict[q_key][key_1][key_2]
                data_dict[q_key].pop(key_1)  # removes key after flattening

        if s_name in self.minionqc_data:
            log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")

        self.minionqc_data[s_name] = data_dict["all"]  # stats for all reads
        self.qfilt_data[s_name] = data_dict["q_filt"]  # stats for q-filtered reads
        self.q_threshold_list.add(q_threshold)  # quality threshold used in this file

    @staticmethod
    def headers_to_use():
        """
        Defines features of columns to be used in multiqc table
        """
        headers = {
            "total.reads": {
                "title": "Total reads",
                "description": "Total number of reads",
                "format": "{:,.0f}",
                "scale": "Greys",
            },
            "total.gigabases": {
                "title": "Total bases (GB)",
                "description": "Total bases",
                "format": "{:,.2f}",
                "scale": "Blues",
            },
            "N50.length": {
                "title": "Reads N50",
                "description": "Minimum read length needed to cover 50% of all reads",
                "format": "{:,.0f}",
                "scale": "Purples",
            },
            "mean.q": {
                "title": "Mean Q score",
                "description": "Mean quality of reads",
                "min": 0,
                "max": 15,
                "format": "{:,.1f}",
                "hidden": True,
                "scale": "Greens",
            },
            "median.q": {
                "title": "Median Q score",
                "description": "Median quality of reads",
                "min": 0,
                "max": 15,
                "format": "{:,.1f}",
                "scale": "Greens",
            },
            "mean.length": {
                "title": "Mean length (bp)",
                "description": "Mean read length",
                "format": "{:,.0f}",
                "hidden": True,
                "scale": "Blues",
            },
            "median.length": {
                "title": "Median length (bp)",
                "description": "Median read length",
                "format": "{:,.0f}",
                "scale": "Blues",
            },
        }

        # Add row ID to avoid duplicates
        for k in headers:
            h_id = re.sub("[^0-9a-zA-Z]+", "_", headers[k]["title"])
            headers[k]["rid"] = f"rid_{h_id}"

        return headers

    def table_qALL(self):
        """Table showing stats for all reads"""

        self.add_section(
            name="Stats: All reads",
            anchor="minionqc-stats-qAll",
            description="MinIONQC statistics for all reads",
            plot=table.plot(
                self.minionqc_data,
                self.headers_to_use(),
                {
                    "namespace": "MinIONQC",
                    "id": "minionqc-stats-qAll-table",
                    "table_title": "MinIONQC Stats: All reads",
                },
            ),
        )

    def table_qfiltered(self):
        """Table showing stats for q-filtered reads"""

        description = "MinIONQC statistics for quality filtered reads. " + "Quailty threshold used: {}.".format(
            ", ".join(list(self.q_threshold_list))
        )
        if len(self.q_threshold_list) > 1:
            description += """
            <div class="alert alert-warning">
              <span class="glyphicon glyphicon-warning-sign"></span>
              <strong>Warning!</strong> More than one quality thresholds were present.
            </div>
            """
            log.warning(
                "More than one quality thresholds were present. Thresholds: {}.".format(
                    ", ".join(list(self.q_threshold_list))
                )
            )

        self.add_section(
            name="Stats: Quality filtered reads",
            anchor="minionqc-stats-qFilt",
            description=description,
            plot=table.plot(
                self.qfilt_data,
                self.headers_to_use(),
                {
                    "namespace": "MinIONQC",
                    "id": "minionqc-stats-qFilt-table",
                    "table_title": "MinIONQC Stats: Quality filtered reads",
                },
            ),
        )

    def plot_readlengths(self):
        pdata = [
            {s_name: d["All reads"]["reads"] for s_name, d in self.minionqc_raw_data.items()},
            {s_name: d["All reads"]["gigabases"] for s_name, d in self.minionqc_raw_data.items()},
        ]
        pconfig = {
            "id": "minionqc_read_lengths",
            "title": "MinIONQC: Output versus read length",
            "categories": True,
            "ylab": "# reads",
            "data_labels": [
                {"name": "All reads: Num reads", "ylab": "# reads"},
                {"name": "All reads: Num gigabases", "ylab": "# gigabases"},
            ],
        }
        for qfilt in list(self.q_threshold_list):
            try:
                pdata.extend(
                    [
                        {s_name: d[qfilt]["reads"] for s_name, d in self.minionqc_raw_data.items()},
                        {s_name: d[qfilt]["gigabases"] for s_name, d in self.minionqc_raw_data.items()},
                    ]
                )
                pconfig["data_labels"].extend(
                    [
                        {"name": f"{qfilt}: Num reads", "ylab": "# reads"},
                        {"name": f"{qfilt}: Num gigabases", "ylab": "# gigabases"},
                    ]
                )
            except KeyError:
                pass
        self.add_section(
            name="Read length output",
            anchor="minionqc-read-length-output",
            description="Number of reads / bp sequenced at given read length thresholds.",
            plot=linegraph.plot(pdata, pconfig=pconfig),
        )



"""
Check that Plotly version is supported, and show an error and exit if it isn't.
"""

import logging
import sys

from packaging import version

logger = logging.getLogger(__name__)


LATEST_SUPPORTED = "5.17"


def check_plotly_version():
    try:
        import plotly
    except ImportError:
        logger.error("ERROR: Could not import Plotly")
        sys.exit(1)
    if version.parse(plotly.__version__) < version.parse(LATEST_SUPPORTED):
        logger.error(
            f"ERROR: found Plotly version {plotly.__version__}, but MultiQC needs at least {LATEST_SUPPORTED}. "
            f'Please upgrade: pip install "plotly>={LATEST_SUPPORTED}"'
        )
        sys.exit(1)


def determine_barplot_height(max_n_samples, max_bars_in_group=1, legend_height=None):
    """
    Used in bar and box plots
    """

    n_bars = max_n_samples
    n_bars *= max_bars_in_group

    def n_bars_to_size(n: int):
        if n >= 30:
            return 15
        if n >= 20:
            return 20
        if n >= 10:
            return 25
        if n >= 5:
            return 30
        return 35

    # Set height to be proportional to the number of samples
    height = n_bars * n_bars_to_size(n_bars)

    if legend_height is not None:
        # expand the plot to fit the legend:
        height = max(height, max(height, legend_height))
        # but not too much - if there are only 2 samples, we don't want the plot to be too high:
        height = min(660, max(height, legend_height))

    height += 140  # Add space for the title and footer

    # now, limit the max and min height (plotly will start to automatically skip
    # some of the ticks on the left when the plot is too high)
    MIN_HEIGHT = 300
    MAX_HEIGHT = 2560
    height = min(MAX_HEIGHT, height)
    height = max(MIN_HEIGHT, height)
    return height



from multiqc import config

read_format = "{:,.1f}"
if config.read_count_multiplier == 1:
    read_format = "{:,.0f}"

base_format = "{:,.1f}"
if config.base_count_multiplier == 1:
    base_format = "{:,.0f}"
elif config.base_count_multiplier == 0.000000001:
    base_format = "{:,.2f}"


class Metric:
    def __init__(
        self,
        id,
        title,
        in_genstats=None,
        in_own_tabl=None,
        unit=None,
        descr=None,
        fmt=None,
        modify=None,
        namespace=None,
        precision=None,
        the_higher_the_worse=False,
    ):
        self.id = id
        self.title = title
        self.in_genstats = in_genstats
        self.in_own_tabl = in_own_tabl
        self.unit = unit
        self.descr = descr
        self.fmt = fmt
        self.modify = modify
        self.namespace = namespace
        self.precision = precision
        self.the_higher_the_worse = the_higher_the_worse


def make_headers(parsed_metric_ids, metrics):
    # Init general stats table
    genstats_headers = {}

    # Init headers for an own separate table
    own_tabl_headers = {}

    for metric in metrics:
        col = dict(
            title=metric.title,
            description=metric.descr,
            min=0,
        )

        # choosing color based on metric namespace, guessing namespace by unit
        if metric.unit == "reads":
            col["scale"] = "Greens"
        elif metric.unit == "bases":
            col["scale"] = "Greys"
        elif metric.unit == "bp":
            col["scale"] = "Purples"
        elif metric.unit == "x" or metric.namespace == "Coverage":
            col["scale"] = "Blues"
            if metric.unit == "%":
                col["scale"] = "RdBu"
                if metric.the_higher_the_worse:
                    col["scale"] = "YlOrRd"
        elif metric.namespace == "Variants":
            col["scale"] = "Purples"

        if metric.id + " pct" in parsed_metric_ids:
            # if % value is available, showing it instead of the number value; the number value will be hidden
            pct_col = dict(
                col,
                description=metric.descr.replace(", {}", "").replace("Number of ", "% of "),
                max=100,
                suffix="%",
            )
            if metric.unit == "reads":
                pct_col["scale"] = "RdYlGn"
            elif metric.unit == "bases":
                pct_col["scale"] = "RdGy"
            elif metric.unit == "x" or metric.namespace == "Coverage":
                pct_col["scale"] = "RdBu"
            if metric.the_higher_the_worse:
                pct_col["scale"] = "YlOrRd"

            if metric.precision is not None:
                pct_col["format"] = "{:,." + str(metric.precision) + "f}"

            if metric.in_own_tabl is not None:
                show = metric.in_own_tabl == "%"
                own_tabl_headers[metric.id + " pct"] = dict(pct_col, hidden=not show)
            if metric.in_genstats is not None and metric.in_own_tabl != "#":
                genstats_headers[metric.id + " pct"] = dict(pct_col, hidden=metric.in_genstats == "hid")

        if metric.unit == "reads":
            col["description"] = col["description"].format(config.read_count_desc)
            col["title"] = config.read_count_prefix + " " + col["title"]
            col["modify"] = lambda x: x * config.read_count_multiplier
            col["shared_key"] = "read_count"
            col["format"] = read_format
        elif metric.unit == "bases":
            col["description"] = col["description"].format(config.base_count_desc)
            col["title"] = config.base_count_prefix + " " + col["title"]
            col["modify"] = lambda x: x * config.base_count_multiplier
            col["shared_key"] = "base_count"
            col["format"] = base_format
        elif metric.unit == "len":
            col["suffix"] = " bp"
            col["format"] = "{:,.0f}"
        elif metric.unit == "x":
            col["suffix"] = " x"
            col["format"] = "{:,.1f}"
        elif metric.unit == "%":
            col["suffix"] = " %"
            col["format"] = "{:,.1f}"
        elif any(
            (metric.descr and metric.descr.startswith(pref))
            for pref in ("Total number of ", "The number of ", "Number of ")
        ):
            col["format"] = "{:,.0f}"
        if metric.precision is not None:
            col["format"] = "{:,." + str(metric.precision) + "f}"

        if metric.in_own_tabl is not None:
            own_tabl_headers[metric.id] = dict(col, hidden=metric.in_own_tabl != "#")
        if metric.in_genstats is not None:
            if metric.in_genstats == "#":
                genstats_headers[metric.id] = dict(col)
            elif metric.in_genstats == "hid" and metric.in_own_tabl == "#":
                genstats_headers[metric.id] = dict(col, hidden=metric.in_genstats == "hid")

    return genstats_headers, own_tabl_headers


def exist_and_number(data, *metrics):
    return all(isinstance(data.get(m, None), int) or isinstance(data.get(m, None), float) for m in metrics)


def check_duplicate_samples(sample_names, logger):
    """Check samples for duplicate names. Warn about found ones."""
    message = ""
    line1 = "\n  {} was built from the following samples:"
    line2 = "\n    {} in {}"
    for clean_sample_name in sample_names:
        if len(sample_names[clean_sample_name]) > 1:
            message += line1.format(clean_sample_name)
            for file_handler in sample_names[clean_sample_name]:
                message += line2.format(file_handler["fn"], file_handler["root"])
    if message:
        message = "\n\nDuplicate sample names were found. The last one overwrites previous data." + message
        logger.debug(message)


def order_headers(headers):
    """Produces a shallow copy with ordered single headers."""
    """
    Collected "order_priority" values are stored as groups to resolve
    conflicts, which may arise if several metrics have the same priority.
    The order of headers within the same priority group is not specified.
    """
    indexes = set()
    ordered_headers = {}
    not_ordered_headers = {}
    for metric in headers:
        if "order_priority" in headers[metric] and isinstance(headers[metric]["order_priority"], (int, float)):
            order_priority = headers[metric]["order_priority"]
            indexes.add(order_priority)
            if order_priority not in ordered_headers:
                ordered_headers[order_priority] = []
            ordered_headers[order_priority].append((metric, headers[metric].copy()))
        else:
            not_ordered_headers[metric] = headers[metric].copy()

    # Convert to list and sort in ascending order.
    if indexes:
        indexes = list(indexes)
        indexes.sort()
    else:
        return headers

    output_headers = {}
    for index in indexes:
        for metric in ordered_headers[index]:
            output_headers[metric[0]] = metric[1]

    output_headers.update(not_ordered_headers)
    return output_headers


# STD_TABLE_CONFIGS contains all standard table configurations from:
# https://github.com/MultiQC/MultiQC/blob/main/docs/plots.md#creating-a-table
STD_TABLE_CONFIGS = [
    "namespace",
    "title",
    "description",
    "max",
    "min",
    "ceiling",
    "floor",
    "minRange",
    "scale",
    "bgcols",
    "colour",
    "suffix",
    "format",
    "cond_formatting_rules",
    "cond_formatting_colours",
    "shared_key",
    "modify",
    "hidden",
    "bars_zero_centrepoint",
]


def clean_headers(headers):
    cleaned_headers = {}
    for metric in headers:
        cleaned_headers[metric] = {
            config: val for config, val in headers[metric].items() if config in STD_TABLE_CONFIGS
        }
    return cleaned_headers


# Used to define module-specific texts for the make_parsing_log_report.
# Comments are used to separate keys. Though empty lines look better, they are removed by the "Black" formatter.
DRAGEN_MODULE_TEXTS = {
    "invalid_file_names": {
        "coverage_metrics": "\n\nThe file names must conform to the following structure:\n"
        "<output-prefix>.<coverage-region-prefix>_coverage_metrics<arbitrary-suffix>.csv\n\n"
        "The following files are not valid:\n",
        # ------------- #
        "overall_mean_cov_metrics": "\n\nThe file names must conform to the following structure:\n"
        "<output-prefix>_overall_mean_cov<arbitrary-suffix>.csv\n\n"
        "The following files are not valid:\n",
    },
    "invalid_file_lines": {
        "coverage_metrics": "\n\nThe lines in files must be:\n"
        "COVERAGE SUMMARY,,<metric>,<value1> or "
        "COVERAGE SUMMARY,,<metric>,<value1>,<value2>\n\n"
        "The following files contain invalid lines:\n",
    },
    "unknown_metrics": {
        "coverage_metrics": "\n\nThe following metrics could not be recognized. "
        "Their headers might be incomplete, hence uninformative and ugly table's columns.\n",
        # ------------- #
        "overall_mean_cov_metrics": "\n\nOnly the 'Average alignment coverage over <file>' metric is supported.\n"
        "The following metrics could not be recognized and are not included in the report.\n",
    },
    "unusual_values": {
        "coverage_metrics": "\n\nAll metrics' values except for int, float and NA are non-standard.\n"
        "The following files contain non-standard values:\n",
        # ------------- #
        "overall_mean_cov_metrics": "\n\nAll metrics' values except for float are non-standard.\n"
        "The following files contain non-standard values:\n",
    },
}


def make_log_report(log_data, logger, module):
    """The only purpose of this function is to create a readable and informative log output
    about found info/warnings/errors, which were found at the time of executing a parser."""

    if "invalid_file_names" in log_data and log_data["invalid_file_names"]:
        if module in DRAGEN_MODULE_TEXTS["invalid_file_names"]:
            log_message = DRAGEN_MODULE_TEXTS["invalid_file_names"][module]
        else:
            log_message = "\n\nThe following files are not valid:\n"

        for root in log_data["invalid_file_names"]:
            log_message += "  " + root + ":\n"
            for file in log_data["invalid_file_names"][root]:
                log_message += "    " + file + "\n"

        logger.warning(log_message + "\n")

    if "invalid_file_lines" in log_data and log_data["invalid_file_lines"]:
        if module in DRAGEN_MODULE_TEXTS["invalid_file_lines"]:
            log_message = DRAGEN_MODULE_TEXTS["invalid_file_lines"][module]
        else:
            log_message = "\n\nThe following files contain invalid lines:\n"

        for root in log_data["invalid_file_lines"]:
            log_message += "  " + root + ":\n"
            for file in log_data["invalid_file_lines"][root]:
                log_message += "    " + file + ":\n"
                for line in log_data["invalid_file_lines"][root][file]:
                    log_message += "      " + line + "\n"

        logger.debug(log_message + "\n")

    if "unknown_metrics" in log_data and log_data["unknown_metrics"]:
        if module in DRAGEN_MODULE_TEXTS["unknown_metrics"]:
            log_message = DRAGEN_MODULE_TEXTS["unknown_metrics"][module]
        else:
            log_message = "\n\nThe following metrics could not be recognized:\n"

        for metric in log_data["unknown_metrics"]:
            log_message += "  " + metric + "\n"

        logger.debug(log_message + "\n")

    if "unusual_values" in log_data and log_data["unusual_values"]:
        if module in DRAGEN_MODULE_TEXTS["unusual_values"]:
            log_message = DRAGEN_MODULE_TEXTS["unusual_values"][module]
        else:
            log_message = "\n\nThe following files contain non-standard values:\n"

        for root in log_data["unusual_values"]:
            log_message += "  " + root + ":\n"
            for file in log_data["unusual_values"][root]:
                log_message += "    " + file + ":\n"
                for metric in log_data["unusual_values"][root][file]:
                    log_message += "      " + metric + " = " + log_data["unusual_values"][root][file][metric] + "\n"

        logger.debug(log_message + "\n")



#!/usr/bin/env python

""" MultiQC functions to plot a bargraph """


import base64
import io
import logging

import math
import os
import random
import sys

from multiqc.utils import config, report, util_functions

logger = logging.getLogger(__name__)

letters = "abcdefghijklmnopqrstuvwxyz"

# Load the template so that we can access its configuration
# Do this lazily to mitigate import-spaghetti when running unit tests
_template_mod = None


def get_template_mod():
    global _template_mod
    if not _template_mod:
        _template_mod = config.avail_templates[config.template].load()
    return _template_mod


def plot(plotdata, plotsamples, pconfig=None):
    """Plot a horizontal bar graph. Expects a 2D dict of sample
    data. Also, can take info about categories. There are quite a
    few variants of how to use this function, see the docs for details.
    :param data: 2D dict, first keys as sample names, then x:y data pairs
                 Can supply a list of dicts and will have buttons to switch
    :param cats: optional list or dict with plot categories
    :param pconfig: optional dict with config key:value pairs
    :return: HTML and JS, ready to be inserted into the page
    """
    if config.plots_force_flat or (
        not config.plots_force_interactive and len(plotsamples[0]) > config.plots_flat_numseries
    ):
        try:
            report.num_mpl_plots += 1
            return matplotlib_bargraph(plotdata, plotsamples, pconfig)
        except Exception as e:
            if config.strict:
                raise
            logger.error("############### Error making MatPlotLib figure! Falling back to HighCharts.")
            logger.debug(e, exc_info=True)
            return highcharts_bargraph(plotdata, plotsamples, pconfig)
    else:
        # Use MatPlotLib to generate static plots if requested
        if config.export_plots:
            try:
                matplotlib_bargraph(plotdata, plotsamples, pconfig)
            except Exception as e:
                logger.error("############### Error making MatPlotLib figure! Plot not exported.")
                logger.debug(e, exc_info=True)
        # Return HTML for HighCharts dynamic plot
        return highcharts_bargraph(plotdata, plotsamples, pconfig)


def highcharts_bargraph(plotdata, plotsamples=None, pconfig=None):
    """
    Build the HTML needed for a HighCharts bar graph. Should be
    called by plot_bargraph, which properly formats input data.
    """
    if pconfig is None:
        pconfig = {}
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_hcplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    html = '<div class="mqc_hcplot_plotgroup">'

    # Counts / Percentages / Log Switches
    if pconfig.get("cpswitch") is not False or pconfig.get("logswitch") is True:
        if pconfig.get("logswitch_active") is True:
            c_active = ""
            p_active = ""
            l_active = "active"
        elif pconfig.get("cpswitch_c_active", True) is True:
            c_active = "active"
            p_active = ""
            l_active = ""
        else:
            c_active = ""
            p_active = "active"
            l_active = ""
            pconfig["stacking"] = "percent"
        c_label = pconfig.get("cpswitch_counts_label", "Counts")
        p_label = pconfig.get("cpswitch_percent_label", "Percentages")
        l_label = pconfig.get("logswitch_label", "Log10")
        html += '<div class="btn-group hc_switch_group"> \n'
        html += '<button class="btn btn-default btn-sm {c_a}" data-action="set_numbers" data-target="{id}" data-ylab="{c_l}">{c_l}</button> \n'.format(
            id=pconfig["id"], c_a=c_active, c_l=c_label
        )
        if pconfig.get("cpswitch", True) is True:
            html += '<button class="btn btn-default btn-sm {p_a}" data-action="set_percent" data-target="{id}" data-ylab="{p_l}">{p_l}</button> \n'.format(
                id=pconfig["id"], p_a=p_active, p_l=p_label
            )
        if pconfig.get("logswitch") is True:
            html += '<button class="btn btn-default btn-sm {l_a}" data-action="set_log" data-target="{id}" data-ylab="{l_l}">{l_l}</button> \n'.format(
                id=pconfig["id"], l_a=l_active, l_l=l_label
            )
            pconfig["reversedStacks"] = True
        html += "</div> "
        if len(plotdata) > 1:
            html += " &nbsp; &nbsp; "

    # Buttons to cycle through different datasets
    if len(plotdata) > 1:
        html += '<div class="btn-group hc_switch_group">\n'
        for k, p in enumerate(plotdata):
            active = "active" if k == 0 else ""
            try:
                name = pconfig["data_labels"][k]["name"]
            except Exception:
                try:
                    name = pconfig["data_labels"][k]
                except Exception:
                    name = k + 1
            try:
                ylab = f"data-ylab=\"{pconfig['data_labels'][k]['ylab']}\""
            except Exception:
                ylab = f'data-ylab="{name}"' if name != k + 1 else ""
            try:
                ymax = f"data-ymax=\"{pconfig['data_labels'][k]['ymax']}\""
            except Exception:
                ymax = ""
            html += '<button class="btn btn-default btn-sm {a}" data-action="set_data" {y} {ym} data-newdata="{k}" data-target="{id}">{n}</button>\n'.format(
                a=active, id=pconfig["id"], n=name, y=ylab, ym=ymax, k=k
            )
        html += "</div>\n\n"

    # Plot HTML
    html += """<div class="hc-plot-wrapper"{height}>
        <div id="{id}" class="hc-plot not_rendered hc-bar-plot"><small>loading..</small></div>
    </div></div>""".format(
        id=pconfig["id"],
        height=f' style="height:{pconfig["height"]}px"' if "height" in pconfig else "",
    )

    report.num_hc_plots += 1

    report.plot_data[pconfig["id"]] = {
        "plot_type": "bar_graph",
        "samples": plotsamples,
        "datasets": plotdata,
        "config": pconfig,
    }

    return html


def matplotlib_bargraph(plotdata, plotsamples, pconfig=None):
    """
    Plot a bargraph with Matplot lib and return a HTML string. Either embeds a base64
    encoded image within HTML or writes the plot and links to it. Should be called by
    plot_bargraph, which properly formats the input data.
    """
    try:
        # Import matplot lib but avoid default X environment
        import matplotlib

        matplotlib.use("Agg")
        import matplotlib.pyplot as plt

        logger.debug(f"Using matplotlib version {matplotlib.__version__}")
    except Exception as e:
        # MatPlotLib can break in a variety of ways. Fake an error message and continue without it if so.
        # The lack of the library will be handled when plots are attempted
        print("##### ERROR! MatPlotLib library could not be loaded!    #####", file=sys.stderr)
        print("##### Flat plots will instead be plotted as interactive #####", file=sys.stderr)
        print(e)

    if pconfig is None:
        pconfig = {}

    # Plot group ID
    if pconfig.get("id") is None:
        pconfig["id"] = "mqc_mplplot_" + "".join(random.sample(letters, 10))

    # Sanitise plot ID and check for duplicates
    pconfig["id"] = report.save_htmlid(pconfig["id"])

    # Individual plot IDs
    pids = []
    for k in range(len(plotdata)):
        try:
            name = pconfig["data_labels"][k]
        except Exception:
            name = k + 1
        pid = f"mqc_{pconfig['id']}_{name}"
        pid = report.save_htmlid(pid, skiplint=True)
        pids.append(pid)

    html = (
        '<p class="text-info"><small><span class="glyphicon glyphicon-picture" aria-hidden="true"></span> '
        + "Flat image plot. Toolbox functions such as highlighting / hiding samples will not work "
        + '(see the <a href="http://multiqc.info/docs/#flat--interactive-plots" target="_blank">docs</a>).</small></p>'
    )
    html += f"<div class=\"mqc_mplplot_plotgroup\" id=\"{pconfig['id']}\">"

    # Counts / Percentages Switch
    if pconfig.get("cpswitch") is not False and not config.simple_output:
        if pconfig.get("cpswitch_c_active", True) is True:
            c_active = "active"
            p_active = ""
        else:
            c_active = ""
            p_active = "active"
            pconfig["stacking"] = "percent"
        c_label = pconfig.get("cpswitch_counts_label", "Counts")
        p_label = pconfig.get("cpswitch_percent_label", "Percentages")
        html += '<div class="btn-group mpl_switch_group mqc_mplplot_bargraph_setcountspcnt"> \n\
            <button class="btn btn-default btn-sm {c_a} counts">{c_l}</button> \n\
            <button class="btn btn-default btn-sm {p_a} pcnt">{p_l}</button> \n\
        </div> '.format(c_a=c_active, p_a=p_active, c_l=c_label, p_l=p_label)
        if len(plotdata) > 1:
            html += " &nbsp; &nbsp; "

    # Buttons to cycle through different datasets
    if len(plotdata) > 1 and not config.simple_output:
        html += '<div class="btn-group mpl_switch_group mqc_mplplot_bargraph_switchds">\n'
        for k, p in enumerate(plotdata):
            pid = pids[k]
            active = "active" if k == 0 else ""
            try:
                name = pconfig["data_labels"][k]
            except Exception:
                name = k + 1
            html += '<button class="btn btn-default btn-sm {a}" data-target="#{pid}">{n}</button>\n'.format(
                a=active, pid=pid, n=name
            )
        html += "</div>\n\n"

    # Go through datasets creating plots
    for pidx, pdata in enumerate(plotdata):
        # Save plot data to file
        fdata = {}
        for d in pdata:
            for didx, dval in enumerate(d["data"]):
                s_name = plotsamples[pidx][didx]
                if s_name not in fdata:
                    fdata[s_name] = dict()
                fdata[s_name][d["name"]] = dval
        if pconfig.get("save_data_file", True):
            util_functions.write_data_file(fdata, pids[pidx])

        # Plot percentage as well as counts
        plot_pcts = [False]
        if pconfig.get("cpswitch") is not False:
            plot_pcts = [False, True]

        # Switch out NaN for 0s so that MatPlotLib doesn't ignore stuff
        for idx, d in enumerate(pdata):
            pdata[idx]["data"] = [x if not math.isnan(x) else 0 for x in d["data"]]

        for plot_pct in plot_pcts:
            # Plot ID
            pid = pids[pidx]
            hide_plot = False
            if plot_pct is True:
                pid = f"{pid}_pc"
                if pconfig.get("cpswitch_c_active", True) is True:
                    hide_plot = True
            else:
                if pconfig.get("cpswitch_c_active", True) is not True:
                    hide_plot = True

            # Set up figure

            # Height has a default, then adjusted by the number of samples
            plt_height = len(plotsamples[pidx]) / 2.3  # Default in inches, empirically determined
            plt_height = max(6, plt_height)  # At least 6" tall
            plt_height = min(30, plt_height)  # Cap at 30" tall

            # Use fixed height if pconfig['height'] is set (convert pixels -> inches)
            if "height" in pconfig:
                # Default interactive height in pixels = 512
                # Not perfect replication, but good enough
                plt_height = 6 * (pconfig["height"] / 512)

            bar_width = 0.8

            fig = plt.figure(figsize=(14, plt_height), frameon=False)
            axes = fig.add_subplot(111)
            y_ind = range(len(plotsamples[pidx]))

            # Count totals for each sample
            if plot_pct is True:
                s_totals = [0 for _ in pdata[0]["data"]]
                for series_idx, d in enumerate(pdata):
                    for sample_idx, v in enumerate(d["data"]):
                        s_totals[sample_idx] += v

            # Plot bars
            dlabels = []
            prev_values = None
            for idx, d in enumerate(pdata):
                # Plot percentages
                values = [x for x in d["data"]]
                if len(values) < len(y_ind):
                    values.extend([0] * (len(y_ind) - len(values)))
                if plot_pct is True:
                    for key, var in enumerate(values):
                        s_total = s_totals[key]
                        if s_total == 0:
                            values[key] = 0
                        else:
                            values[key] = (float(var + 0.0) / float(s_total)) * 100

                # Get offset for stacked bars
                if idx == 0:
                    prevdata = [0] * len(plotsamples[pidx])
                else:
                    for i, p in enumerate(prevdata):
                        prevdata[i] += prev_values[i]
                # Save the name of this series
                dlabels.append(d["name"])
                # Add the series of bars to the plot
                axes.barh(
                    y_ind,
                    values,
                    bar_width,
                    left=prevdata,
                    color=d["color"],
                    align="center",
                    linewidth=pconfig.get("borderWidth", 0),
                )
                prev_values = values

            # Tidy up axes
            axes.tick_params(
                labelsize=pconfig.get("labelSize", 8), direction="out", left=False, right=False, top=False, bottom=False
            )
            axes.set_xlabel(pconfig.get("ylab", ""))  # I know, I should fix the fact that the config is switched
            axes.set_ylabel(pconfig.get("xlab", ""))
            axes.set_yticks(y_ind)  # Specify where to put the labels
            axes.set_yticklabels(plotsamples[pidx])  # Set y axis sample name labels
            axes.set_ylim((-0.5, len(y_ind) - 0.5))  # Reduce padding around plot area
            if plot_pct is True:
                axes.set_xlim((0, 100))
                # Add percent symbols
                vals = axes.get_xticks()
                axes.set_xticks(axes.get_xticks())
                axes.set_xticklabels([f"{x:.0f}%" for x in vals])
            else:
                default_xlimits = axes.get_xlim()
                axes.set_xlim((pconfig.get("ymin", default_xlimits[0]), pconfig.get("ymax", default_xlimits[1])))
            if "title" in pconfig:
                top_gap = 1 + (0.5 / plt_height)
                plt.text(
                    0.5, top_gap, pconfig["title"], horizontalalignment="center", fontsize=16, transform=axes.transAxes
                )
            axes.grid(True, zorder=0, which="both", axis="x", linestyle="-", color="#dedede", linewidth=1)
            axes.set_axisbelow(True)
            axes.spines["right"].set_visible(False)
            axes.spines["top"].set_visible(False)
            axes.spines["bottom"].set_visible(False)
            axes.spines["left"].set_visible(False)
            plt.gca().invert_yaxis()  # y-axis is reverse sorted otherwise

            # Hide some labels if we have a lot of samples
            show_nth = max(1, math.ceil(len(pdata[0]["data"]) / 150))
            for idx, label in enumerate(axes.get_yticklabels()):
                if idx % show_nth != 0:
                    label.set_visible(False)

            # Legend
            bottom_gap = -1 * (1 - ((plt_height - 1.5) / plt_height))
            lgd = axes.legend(
                dlabels,
                loc="lower center",
                bbox_to_anchor=(0, bottom_gap, 1, 0.102),
                ncol=5,
                mode="expand",
                fontsize=pconfig.get("labelSize", 8),
                frameon=False,
            )

            # Should this plot be hidden on report load?
            hidediv = ""
            if pidx > 0 or hide_plot:
                hidediv = ' style="display:none;"'

            # Save the plot to the data directory if export is requested
            if config.export_plots:
                for fformat in config.export_plot_formats:
                    # Make the directory if it doesn't already exist
                    plot_dir = os.path.join(config.plots_dir, fformat)
                    if not os.path.exists(plot_dir):
                        os.makedirs(plot_dir)
                    # Save the plot
                    plot_fn = os.path.join(plot_dir, f"{pid}.{fformat}")
                    fig.savefig(plot_fn, format=fformat, bbox_extra_artists=(lgd,), bbox_inches="tight")

            # Output the figure to a base64 encoded string
            if getattr(get_template_mod(), "base64_plots", True) is True:
                img_buffer = io.BytesIO()
                fig.savefig(img_buffer, format="png", bbox_inches="tight")
                b64_img = base64.b64encode(img_buffer.getvalue()).decode("utf8")
                img_buffer.close()
                html += '<div class="mqc_mplplot" id="{}"{}><img src="data:image/png;base64,{}" /></div>'.format(
                    pid, hidediv, b64_img
                )

            # Link to the saved image
            else:
                plot_relpath = os.path.join(config.plots_dir_name, "png", f"{pid}.png")
                html += f'<div class="mqc_mplplot" id="{pid}"{hidediv}><img src="{plot_relpath}" /></div>'

            plt.close(fig)

    # Close wrapping div
    html += "</div>"

    return html



""" MultiQC module to parse output from Slamdunk """


import logging
import re

from multiqc import config
from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph, scatter, table

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"# slamdunk summary v([\d\.]+)"


class MultiqcModule(BaseMultiqcModule):
    """
    Slamdunk module class, parses slamdunk logs.
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Slamdunk",
            anchor="slamdunk",
            href="http://t-neumann.github.io/slamdunk/",
            info="is a tool to analyze SLAMSeq data.",
            doi="10.1186/s12859-019-2849-7",
        )

        num_reports = 0

        self.plot_cols = [
            "#fdbf6f",
            "#2171B5",
            "#6BAED6",
            "#C6DBEF",
            "#74C476",
            "#C7E9C0",
            "#D9D9D9",
            "#969696",
            "#525252",
            "#DADAEB",
            "#9E9AC8",
            "#6A51A3",
        ]

        # Summary Reports
        self.slamdunk_data = dict()
        for f in self.find_log_files("slamdunk/summary", filehandles=True):
            self.parseSummary(f)
        self.slamdunk_data = self.ignore_samples(self.slamdunk_data)
        if len(self.slamdunk_data) > 0:
            self.slamdunkGeneralStatsTable()
            self.slamdunkFilterStatsTable()
            log.debug(f"Found {len(self.slamdunk_data)} summary reports")
            num_reports = max(num_reports, len(self.slamdunk_data))

        # PCA Plots
        self.PCA_data = dict()
        for f in self.find_log_files("slamdunk/PCA", filehandles=True):
            self.parsePCA(f)
        self.PCA_data = self.ignore_samples(self.PCA_data)
        if len(self.PCA_data) > 0:
            self.slamdunkPCAPlot()
            log.debug(f"Found {len(self.PCA_data)} PCA plots")
            num_reports = max(num_reports, len(self.PCA_data))

        # UTR Rate reports
        self.utrates_data = dict()
        for f in self.find_log_files("slamdunk/utrrates", filehandles=True):
            self.parseUtrRates(f)
        self.utrates_data = self.ignore_samples(self.utrates_data)
        if len(self.utrates_data) > 0:
            self.write_data_file(self.utrates_data, "multiqc_slamdunk_utrrates")
            self.slamdunkUtrRatesPlot()
            log.debug(f"Found {len(self.utrates_data)} UTR rate reports")
            num_reports = max(num_reports, len(self.utrates_data))

        # Read rate reports
        self.rates_data_plus = dict()
        self.rates_data_minus = dict()
        for f in self.find_log_files("slamdunk/rates", filehandles=True):
            self.parseSlamdunkRates(f)
        self.rates_data_plus = self.ignore_samples(self.rates_data_plus)
        self.rates_data_minus = self.ignore_samples(self.rates_data_minus)
        if len(self.rates_data_plus) > 0:
            self.write_data_file(self.rates_data_plus, "multiqc_slamdunk_readrates_plus")
            self.write_data_file(self.rates_data_minus, "multiqc_slamdunk_readrates_minus")
            self.slamdunkOverallRatesPlot()
            log.debug(f"Found {len(self.rates_data_plus)} read rate reports")
            num_reports = max(num_reports, len(self.rates_data_plus))

        # TCP error read rate
        self.nontc_per_readpos_plus = dict()
        self.nontc_per_readpos_minus = dict()
        self.tc_per_readpos_plus = dict()
        self.tc_per_readpos_minus = dict()
        for f in self.find_log_files("slamdunk/tcperreadpos", filehandles=True):
            self.parseSlamdunkTCPerReadpos(f)
        self.nontc_per_readpos_plus = self.ignore_samples(self.nontc_per_readpos_plus)
        self.nontc_per_readpos_minus = self.ignore_samples(self.nontc_per_readpos_minus)
        self.tc_per_readpos_plus = self.ignore_samples(self.tc_per_readpos_plus)
        self.tc_per_readpos_minus = self.ignore_samples(self.tc_per_readpos_minus)
        if len(self.tc_per_readpos_plus) > 0:
            self.write_data_file(self.tc_per_readpos_plus, "multiqc_slamdunk_tcperreadpos_plus")
            self.write_data_file(self.nontc_per_readpos_plus, "multiqc_slamdunk_nontcperreadpos_plus")
            self.write_data_file(self.tc_per_readpos_minus, "multiqc_slamdunk_tcperreadpos_minus")
            self.write_data_file(self.nontc_per_readpos_minus, "multiqc_slamdunk_nontcperreadpos_minus")
            self.slamdunkTcPerReadPosPlot()
            log.debug(f"Found {len(self.tc_per_readpos_plus)} TCP error read rate reports")
            num_reports = max(num_reports, len(self.tc_per_readpos_plus))

        # Non-TCP error read rate
        self.nontc_per_utrpos_plus = dict()
        self.nontc_per_utrpos_minus = dict()
        self.tc_per_utrpos_plus = dict()
        self.tc_per_utrpos_minus = dict()
        for f in self.find_log_files("slamdunk/tcperutrpos", filehandles=True):
            self.parseSlamdunkTCPerUtrpos(f)
        self.nontc_per_utrpos_plus = self.ignore_samples(self.nontc_per_utrpos_plus)
        self.nontc_per_utrpos_minus = self.ignore_samples(self.nontc_per_utrpos_minus)
        self.tc_per_utrpos_plus = self.ignore_samples(self.tc_per_utrpos_plus)
        self.tc_per_utrpos_minus = self.ignore_samples(self.tc_per_utrpos_minus)
        if len(self.nontc_per_utrpos_plus) > 0:
            self.write_data_file(self.tc_per_utrpos_plus, "multiqc_slamdunk_tcperutrpos_plus")
            self.write_data_file(self.nontc_per_utrpos_plus, "multiqc_slamdunk_nontcperutrpos_plus")
            self.write_data_file(self.tc_per_utrpos_minus, "multiqc_slamdunk_tcperutrpos_minus")
            self.write_data_file(self.nontc_per_utrpos_minus, "multiqc_slamdunk_nontcperutrpos_minus")
            self.slamdunkTcPerUTRPosPlot()
            log.debug(f"Found {len(self.nontc_per_utrpos_plus)} non TCP error read rate reports")
            num_reports = max(num_reports, len(self.nontc_per_utrpos_plus))

        if num_reports == 0:
            raise ModuleNoSamplesFound
        else:
            log.info(f"Found {num_reports} reports")

    def parsePCA(self, f):
        # Skip header
        next(f["f"])

        for line in f["f"]:
            fields = line.rstrip().split("\t")

            sample = self.clean_s_name(fields[0], f)
            PC1 = fields[1]
            PC2 = fields[2]

            self.PCA_data[sample] = [{"x": float(PC1), "y": float(PC2)}]

    def parseUtrRates(self, f):
        # Skip comment line #
        next(f["f"])

        # Read median header
        line = next(f["f"])

        if "Conversions=" in line:
            sample = f["s_name"]
            self.utrates_data[sample] = dict()

            conversions = re.sub(".*Conversions=", "", line.rstrip()).split(",")

            for conversion in conversions:
                type, value = conversion.split(":")
                self.utrates_data[sample][type] = float(value)

        else:
            log.warning("Malformed UTR rates header. Conversion rates per UTR plot will be affected.")

    def parseSlamdunkRates(self, f):
        sample = f["s_name"]

        # Skip comment line #
        next(f["f"])

        bases = next(f["f"]).rstrip().split("\t")

        baseDict = {}
        order = {}

        for i in range(1, len(bases)):
            order[i] = bases[i]

        for line in f["f"]:
            values = line.rstrip().split("\t")
            base = values[0]
            baseDict[base] = {}

            for i in range(1, len(values)):
                baseDict[base][order[i]] = int(values[i])

        divisor = {}

        for fromBase in baseDict:
            for toBase in baseDict[fromBase]:
                if toBase.islower():
                    if fromBase.lower() not in divisor:
                        divisor[fromBase.lower()] = 0
                    divisor[fromBase.lower()] += baseDict[fromBase][toBase]
                else:
                    if fromBase not in divisor:
                        divisor[fromBase] = 0
                    divisor[fromBase] += baseDict[fromBase][toBase]

        for fromBase in baseDict:
            for toBase in baseDict[fromBase]:
                if toBase.islower():
                    if divisor[fromBase.lower()] > 0:
                        baseDict[fromBase][toBase] = baseDict[fromBase][toBase] / float(divisor[fromBase.lower()]) * 100
                    else:
                        baseDict[fromBase][toBase] = 0.0
                else:
                    if divisor[fromBase] > 0:
                        baseDict[fromBase][toBase] = baseDict[fromBase][toBase] / float(divisor[fromBase]) * 100
                    else:
                        baseDict[fromBase][toBase] = 0.0

        self.rates_data_plus[sample] = {}
        self.rates_data_minus[sample] = {}

        for fromBase in baseDict:
            for toBase in baseDict[fromBase]:
                if fromBase != "N" and toBase.upper() != "N" and fromBase != toBase.upper():
                    if toBase.islower():
                        self.rates_data_minus[sample][fromBase + ">" + toBase.upper()] = baseDict[fromBase][toBase]
                    else:
                        self.rates_data_plus[sample][fromBase + ">" + toBase] = baseDict[fromBase][toBase]

    def parseSlamdunkTCPerReadpos(self, f):
        sample = f["s_name"]

        # Skip comment line #
        next(f["f"])

        self.nontc_per_readpos_plus[sample] = {}
        self.nontc_per_readpos_minus[sample] = {}

        self.tc_per_readpos_plus[sample] = {}
        self.tc_per_readpos_minus[sample] = {}

        pos = 1

        for line in f["f"]:
            values = line.rstrip().split("\t")
            if int(values[4]) > 0:
                self.nontc_per_readpos_plus[sample][pos] = float(int(values[0])) / int(values[4]) * 100
                self.tc_per_readpos_plus[sample][pos] = float(int(values[2])) / int(values[4]) * 100
            else:
                self.nontc_per_readpos_plus[sample][pos] = 0
                self.tc_per_readpos_plus[sample][pos] = 0

            if int(values[5]) > 0:
                self.nontc_per_readpos_minus[sample][pos] = float(int(values[1])) / int(values[5]) * 100
                self.tc_per_readpos_minus[sample][pos] = float(int(values[3])) / int(values[5]) * 100
            else:
                self.nontc_per_readpos_minus[sample][pos] = 0
                self.tc_per_readpos_minus[sample][pos] = 0

            pos += 1

    def parseSlamdunkTCPerUtrpos(self, f):
        sample = f["s_name"]

        # Skip comment line #
        next(f["f"])

        self.nontc_per_utrpos_plus[sample] = {}
        self.nontc_per_utrpos_minus[sample] = {}

        self.tc_per_utrpos_plus[sample] = {}
        self.tc_per_utrpos_minus[sample] = {}

        pos = 1

        for line in f["f"]:
            values = line.rstrip().split("\t")
            if int(values[4]) > 0:
                self.nontc_per_utrpos_plus[sample][pos] = float(int(values[0])) / int(values[4]) * 100
                self.tc_per_utrpos_plus[sample][pos] = float(int(values[2])) / int(values[4]) * 100
            else:
                self.nontc_per_utrpos_plus[sample][pos] = 0
                self.tc_per_utrpos_plus[sample][pos] = 0

            if int(values[5]) > 0:
                self.nontc_per_utrpos_minus[sample][pos] = float(int(values[1])) / int(values[5]) * 100
                self.tc_per_utrpos_minus[sample][pos] = float(int(values[3])) / int(values[5]) * 100
            else:
                self.nontc_per_utrpos_minus[sample][pos] = 0
                self.tc_per_utrpos_minus[sample][pos] = 0

            pos += 1

    def parseSummary(self, f):
        # Parse version form first line
        first = next(f["f"])
        version = None
        match = re.search(VERSION_REGEX, first)
        if match:
            version = match.group(1)

        # Skip header line "FileName..."
        columnCount = next(f["f"]).count("\t") + 1

        for line in f["f"]:
            fields = line.rstrip().split("\t")
            s_name = self.clean_s_name(fields[0], f)
            self.slamdunk_data[s_name] = dict()
            self.slamdunk_data[s_name]["sequenced"] = int(fields[4])
            self.slamdunk_data[s_name]["mapped"] = int(fields[5])
            # self.slamdunk_data[s_name]['deduplicated'] = int(fields[6])
            self.slamdunk_data[s_name]["mqfiltered"] = int(fields[7])
            self.slamdunk_data[s_name]["idfiltered"] = int(fields[8])
            self.slamdunk_data[s_name]["nmfiltered"] = int(fields[9])
            self.slamdunk_data[s_name]["multimapper"] = int(fields[10])
            self.slamdunk_data[s_name]["retained"] = int(fields[11])

            # Additional Count Column found in Table
            if columnCount == 14:
                self.slamdunk_data[s_name]["counted"] = int(fields[12])

        for s_name in self.slamdunk_data.keys():
            self.add_software_version(version, s_name)
            self.add_data_source(f, s_name=s_name)

    def slamdunkGeneralStatsTable(self):
        """Take the parsed summary stats from Slamdunk and add it to the
        basic stats table at the top of the report"""

        headers = {
            "counted": {
                "title": f"{config.read_count_prefix} Counted",
                "description": f"# reads counted within 3'UTRs ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "scale": "YlGn",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "retained": {
                "title": f"{config.read_count_prefix} Retained",
                "description": f"# retained reads after filtering ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "scale": "YlGn",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "mapped": {
                "title": f"{config.read_count_prefix} Mapped",
                "description": f"# mapped reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "scale": "YlGn",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "sequenced": {
                "title": f"{config.read_count_prefix} Sequenced",
                "description": f"# sequenced reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "scale": "YlGn",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
        }

        self.general_stats_addcols(self.slamdunk_data, headers)

    def slamdunkFilterStatsTable(self):
        """Take the parsed filter stats from Slamdunk and add it to a separate table"""

        headers = {
            "mapped": {
                "namespace": "Slamdunk",
                "title": f"{config.read_count_prefix} Mapped",
                "description": f"# mapped reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "suffix": config.read_count_prefix,
                "scale": "YlGn",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "multimapper": {
                "namespace": "Slamdunk",
                "title": f"{config.read_count_prefix} Multimap-Filtered",
                "description": f"# multimap-filtered reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "suffix": config.read_count_prefix,
                "scale": "OrRd",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "nmfiltered": {
                "namespace": "Slamdunk",
                "title": f"{config.read_count_prefix} NM-Filtered",
                "description": f"# NM-filtered reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "suffix": config.read_count_prefix,
                "scale": "OrRd",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "idfiltered": {
                "namespace": "Slamdunk",
                "title": f"{config.read_count_prefix} Identity-Filtered",
                "description": f"# identity-filtered reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "suffix": config.read_count_prefix,
                "scale": "OrRd",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
            "mqfiltered": {
                "namespace": "Slamdunk",
                "title": f"{config.read_count_prefix} MQ-Filtered",
                "description": f"# MQ-filtered reads ({config.read_count_desc})",
                "shared_key": "read_count",
                "min": 0,
                "format": "{:,.2f}",
                "suffix": config.read_count_prefix,
                "scale": "OrRd",
                "modify": lambda x: float(x) * config.read_count_multiplier,
            },
        }
        pconfig = {
            "id": "slamdunk_filtering_table",
            "min": 0,
            "title": "Slamdunk: Filtering statistics",
        }

        self.add_section(
            name="Filter statistics",
            anchor="slamdunk_filtering",
            description="This table shows the number of reads filtered with each filter criterion during filtering phase of slamdunk.",
            plot=table.plot(self.slamdunk_data, headers, pconfig),
        )

    def slamdunkOverallRatesPlot(self):
        """Generate the overall rates plot"""

        pconfig = {
            "id": "overallratesplot",
            "title": "Slamdunk: Overall conversion rates in reads",
            "cpswitch": False,
            "cpswitch_c_active": False,
            "ylab": "Number of reads",
            "stacking": "normal",
            "tt_decimals": 2,
            "tt_suffix": "%",
            "tt_percentages": False,
            "hide_zero_cats": False,
            "data_labels": [
                "Plus Strand +",
                "Minus Strand -",
            ],
        }

        cats = [dict(), dict()]
        keys = [
            ["T>C", "A>T", "A>G", "A>C", "T>A", "T>G", "G>A", "G>T", "G>C", "C>A", "C>T", "C>G"],
            ["A>G", "A>T", "A>C", "T>A", "T>G", "T>C", "G>A", "G>T", "G>C", "C>A", "C>T", "C>G"],
        ]
        for i, k in enumerate(keys):
            for j, v in enumerate(k):
                cats[i][v] = {"color": self.plot_cols[j]}

        self.add_section(
            name="Conversion rates per read",
            anchor="slamdunk_overall_rates",
            description="""This plot shows the individual conversion rates over all reads.
                        It shows these conversion rates strand-specific: This means for a properly labeled
                        sample you would see a T&gt;C excess on the plus-strand and an A&gt;G excess on the minus strand
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#rates" target="_blank">slamdunk docs</a>).""",
            plot=bargraph.plot([self.rates_data_plus, self.rates_data_minus], cats, pconfig),
        )

    def slamdunkUtrRatesPlot(self):
        """Generate the UTR rates plot"""

        cats = dict()
        keys = ["T>C", "A>T", "A>G", "A>C", "T>A", "T>G", "G>A", "G>T", "G>C", "C>A", "C>T", "C>G"]
        for i, v in enumerate(keys):
            cats[v] = {"color": self.plot_cols[i]}

        pconfig = {
            "id": "slamdunk_utrratesplot",
            "title": "Slamdunk: Overall conversion rates per UTR",
            "cpswitch": False,
            "cpswitch_c_active": False,
            "ylab": "Number of conversions",
            "stacking": "normal",
            "tt_decimals": 2,
            "tt_suffix": "%",
            "tt_percentages": False,
            "hide_zero_cats": False,
        }

        self.add_section(
            name="Conversion rates per UTR",
            anchor="slamdunk_utr_rates",
            description="""This plot shows the individual conversion rates for all UTRs
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#utrrates" target="_blank">slamdunk docs</a>).""",
            plot=bargraph.plot(self.utrates_data, cats, pconfig),
        )

    def slamdunkPCAPlot(self):
        """Generate the PCA plots"""

        pconfig = {
            "id": "slamdunk_pca",
            "title": "Slamdunk: PCA",
            "xlab": "PC1",
            "ylab": "PC2",
            "tt_label": "PC1 {point.x:.2f}: PC2 {point.y:.2f}",
        }

        self.add_section(
            name="PCA (T&gt;C based)",
            anchor="slamdunk_PCA",
            description="""This plot shows the principal components of samples based
                        on the distribution of reads with T&gt;C conversions within UTRs
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#summary" target="_blank">slamdunk docs</a>).""",
            plot=scatter.plot(self.PCA_data, pconfig),
        )

    def slamdunkTcPerReadPosPlot(self):
        """Generate the tc per read pos plots"""

        pconfig_nontc = {
            "id": "slamdunk_nontcperreadpos_plot",
            "title": "Slamdunk: Non-T>C mismatches over reads",
            "ylab": "Percent mismatches %",
            "xlab": "Position in read",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Pos {point.x}</b>: {point.y:.2f} %",
            "data_labels": [
                {"name": "Forward reads +", "ylab": "Percent mismatches %"},
                {"name": "Reverse reads -", "ylab": "Percent mismatches %"},
            ],
        }

        pconfig_tc = {
            "id": "slamdunk_tcperreadpos_plot",
            "title": "Slamdunk: T>C conversions over reads",
            "ylab": "Percent converted %",
            "xlab": "Position in read",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Pos {point.x}</b>: {point.y:.2f} %",
            "data_labels": [
                {"name": "Forward reads +", "ylab": "Percent converted %"},
                {"name": "Reverse reads -", "ylab": "Percent converted %"},
            ],
        }

        self.add_section(
            name="Non T&gt;C mismatches over read positions",
            anchor="slamdunk_nontcperreadpos",
            description="""This plot shows the distribution of non T&gt;C mismatches across read positions
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#tcperreadpos" target="_blank">slamdunk docs</a>).""",
            plot=linegraph.plot([self.nontc_per_readpos_plus, self.nontc_per_readpos_minus], pconfig_nontc),
        )

        self.add_section(
            name="T&gt;C conversions over read positions",
            anchor="slamdunk_tcperreadpos",
            description="""This plot shows the distribution of T&gt;C conversions across read positions
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#tcperreadpos" target="_blank">slamdunk docs</a>).""",
            plot=linegraph.plot([self.tc_per_readpos_plus, self.tc_per_readpos_minus], pconfig_tc),
        )

    def slamdunkTcPerUTRPosPlot(self):
        """Generate the tc per UTR pos plots"""

        pconfig_nontc = {
            "id": "slamdunk_slamdunk_nontcperutrpos_plot",
            "title": "Slamdunk: Non-T>C mutations over 3' UTR ends",
            "ylab": "Percent mismatches %",
            "xlab": "Position in the static last 250bp window of 3' UTR",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Pos {point.x}</b>: {point.y:.2f} %",
            "data_labels": [
                {"name": "UTRs on plus strand", "ylab": "Percent mismatches %"},
                {"name": "UTRs on minus strand", "ylab": "Percent mismatches %"},
            ],
        }

        pconfig_tc = {
            "id": "slamdunk_slamdunk_tcperutrpos_plot",
            "title": "Slamdunk: T>C conversions over 3' UTR ends",
            "ylab": "Percent converted %",
            "xlab": "Position in the static last 250bp window of 3' UTR",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>Pos {point.x}</b>: {point.y:.2f} %",
            "data_labels": [
                {"name": "UTRs on plus strand", "ylab": "Percent converted %"},
                {"name": "UTRs on minus strand", "ylab": "Percent converted %"},
            ],
        }

        self.add_section(
            name="Non T&gt;C mismatches over UTR positions",
            anchor="slamdunk_nontcperutrpos",
            description="""This plot shows the distribution of non T&gt;C mismatches across UTR positions for the last 250 bp from the 3\' UTR end
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#tcperutrpos" target="_blank">slamdunk docs</a>).""",
            plot=linegraph.plot([self.nontc_per_utrpos_plus, self.nontc_per_utrpos_minus], pconfig_nontc),
        )

        self.add_section(
            name="T&gt;C conversions over UTR positions",
            anchor="tcperutrpos",
            description="""This plot shows the distribution of T&gt;C conversions across UTR positions for the last 250 bp from the 3\' UTR end
                        (see the <a href="http://t-neumann.github.io/slamdunk/docs.html#tcperutrpos" target="_blank">slamdunk docs</a>).""",
            plot=linegraph.plot([self.tc_per_utrpos_plus, self.tc_per_utrpos_minus], pconfig_tc),
        )



""" MultiQC module to parse log output from xengsort classify """

from collections import defaultdict

import logging
from typing import Dict

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="xengsort",
            anchor="xengsort",
            href="https://gitlab.com/genomeinformatics/xengsort",
            info="is a fast xenograft read sorter based on space-efficient k-mer hashing",
            doi="doi.org/10.4230/LIPIcs.WABI.2020.4",
        )

        # Find and load any Xenome reports
        self.percents = dict()
        self.counts = dict()
        for f in self.find_log_files("xengsort"):
            self._parse_log(f)

        # Filter to strip out ignored sample names
        self.percents = self.ignore_samples(self.percents)
        self.counts = self.ignore_samples(self.counts)
        if len(self.percents) == 0:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.counts)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.percents, "multiqc_xengsort_percents")
        self.write_data_file(self.counts, "multiqc_xengsort_counts")

        self._build_table()
        self._build_plot()

    def _parse_log(self, f):
        lines = iter(f["contents_lines"])
        for line in iter(lines):
            if "\t" in line:
                fields = line.strip().split("\t")
                if set(fields) == {"prefix", "host", "graft", "ambiguous", "both", "neither"}:
                    values = next(lines).strip().split("\t")
                    data = dict(zip(fields, values))
                    s_name = data.pop("prefix")
                    f["s_name"] = s_name
                    data = {k: int(v) for k, v in data.items()}
                    percents = {k: v / sum(data.values()) * 100 for k, v in data.items()}

                    if s_name in self.counts:
                        log.debug(f"Duplicate sample name found! Overwriting: {s_name}")
                    self.add_data_source(f, s_name)
                    self.counts[s_name] = data
                    self.percents[s_name] = percents
                    break

    def _build_table(self):
        """
        Prepare headers and data for a table. Add a section with a table,
        and add a few columns into the general stats.
        """
        headers: Dict[str, Dict] = {}
        table_data = defaultdict(dict)

        scale_by_cls = {
            "graft": "Blues",
            "host": "Reds",
            "both": "Purples",
            "ambiguous": "Greys",
            "neither": "Greys",
        }
        for sn, data in self.percents.items():
            for cls in ["graft", "host", "ambiguous", "both", "neither"]:
                table_data[sn][f"{cls}_reads_pct"] = data.get(cls)
                headers[f"{cls}_reads_pct"] = {
                    "rid": f"xengsort_{cls}_reads_pct",  # to make the ID unique from xenome
                    "title": f"{cls.capitalize()} reads",
                    "description": f"share of {cls} reads in the sample",
                    "min": 0,
                    "suffix": "%",
                    "scale": scale_by_cls[cls],
                    "format": "{:,.2f}",
                    "hidden": cls in ["both", "neither", "ambiguous"],
                }
        self.general_stats_addcols(table_data, headers)
        detail_headers = headers.copy()
        for metric in headers:
            detail_headers[metric]["hidden"] = False

        for sn, data in self.counts.items():
            for cls in ["graft", "host", "ambiguous", "both", "neither"]:
                table_data[sn][f"{cls}_reads_cnt"] = data.get(cls)
                detail_headers[f"{cls}_reads_cnt"] = {
                    "rid": f"xengsort_{cls}_reads_cnt",  # to make the ID unique from xenome
                    "title": f"{cls.capitalize()} reads",
                    "description": f"number of {cls} reads in the sample",
                    "min": 0,
                    "scale": scale_by_cls.get(cls),
                    "format": "{:,d}",
                    "hidden": True,
                }

        self.add_section(
            name="Summary table",
            anchor="xengsort-summary-table-section",
            plot=table.plot(
                table_data,
                detail_headers,
                {
                    "id": "xengsort-summary-table",
                    "namespace": "Xengsort",
                    "title": "Xengsort: Summary",
                },
            ),
        )

    def _build_plot(self):
        """
        Create two bar plots: based on summary and detail data.
        """
        cats = {
            "graft": {"name": "Graft", "color": "#377eb8"},  # blue
            "host": {"name": "Host", "color": "#e41a1c"},  # red
            "both": {"name": "Both", "color": "#984ea3"},  # purple
            "ambiguous": {"name": "Ambiguous", "color": "#616161"},  # grey
            "neither": {"name": "Neither", "color": "#b3b3b3"},  # light grey
        }
        self.add_section(
            description=f"This plot shows the number of reads classified by {self.name}",
            helptext="""
            There are 5 possible categories:  
            * **Graft**: reads found in graft species, e.g. human
            * **Host**: reads found in host species, e.g. mouse
            * **Both**: reads found in either of the species
            * **Neither**: reads was found in neither of the species
            * **Ambiguous**: reads origin could not be adequately determined.  
            """,
            name="Summary classification",
            anchor="xengsort_summary_bar_plot_section",
            plot=bargraph.plot(
                self.counts,
                cats,
                {
                    "id": "xengsort_summary_bar_plot",
                    "title": f"{self.name}: summary classification",
                    "ylab": "# Reads",
                    "cpswitch_counts_label": "Number of reads",
                    "cpswitch_c_active": False,
                    "logswitch": True,
                },
            ),
        )



""" MultiQC module to parse output from RSEM/rsem-calculate-expression """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, linegraph
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    """
    RSEM module class, parses .cnt file .
    """

    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Rsem",
            anchor="rsem",
            href="https://deweylab.github.io/RSEM/",
            info="RSEM (RNA-Seq by Expectation-Maximization) is a software package for"
            "estimating gene and isoform expression levels from RNA-Seq data.",
            doi="10.1186/1471-2105-12-323",
        )

        self.rsem_mapped_data = dict()
        self.rsem_multimapping_data = dict()

        # Find and load any count file
        for f in self.find_log_files("rsem"):
            self.parse_rsem_report(f)

        # Filter to strip out ignored sample names
        self.rsem_mapped_data = self.ignore_samples(self.rsem_mapped_data)

        if len(self.rsem_mapped_data) == 0:
            raise ModuleNoSamplesFound

        log.info(f"Found {len(self.rsem_mapped_data)} reports")

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # Write parsed report data to a file
        self.write_data_file(self.rsem_mapped_data, "multiqc_rsem")

        # Basic Stats Table
        self.rsem_stats_table()

        # Assignment bar plot
        self.rsem_mapped_reads_plot()

        # Multimapping line plot
        self.rsem_multimapping_plot()

    def parse_rsem_report(self, f):
        """Parse the rsem cnt stat file.
        Description of cnt file found : https://github.com/deweylab/RSEM/blob/master/cnt_file_description.txt
        """
        data = dict()
        multimapping_hist = dict()
        in_hist = False
        for line in f["f"].splitlines():
            s = line.split()
            if len(s) > 3:
                # Line: N0 N1 N2 N_tot
                # N0, number of unalignable reads;
                # N1, number of alignable reads; (nUnique + nMulti)
                # N2, number of filtered reads due to too many alignments
                # N_tot = N0 + N1 + N2
                data["Unalignable"] = int(s[0])
                data["Alignable"] = int(s[1])
                data["Filtered"] = int(s[2])
                data["Total"] = int(s[3])
                try:
                    data["alignable_percent"] = (float(s[1]) / float(s[3])) * 100.0
                except ZeroDivisionError:
                    data["alignable_percent"] = 0
            elif len(s) == 3:
                # Line: nUnique nMulti nUncertain
                # nUnique, number of reads aligned uniquely to a gene
                # nMulti, number of reads aligned to multiple genes; nUnique + nMulti = N1;
                # nUncertain, number of reads aligned to multiple locations in the given reference sequences, which include isoform-level multi-mapping reads
                data["Unique"] = int(s[0])
                data["Multi"] = int(s[1])
                data["Uncertain"] = int(s[2])
            elif len(s) == 2:
                # Number of gene alignments and read count
                if in_hist or int(s[0]) == 0:
                    in_hist = True
                    try:
                        multimapping_hist[int(s[0])] = int(s[1])
                    except ValueError:
                        pass
            else:
                break
        try:
            assert data["Unique"] + data["Multi"] == data["Alignable"]
        except AssertionError:
            log.warning(f"Unique + Multimapping read counts != alignable reads! '{f['fn']}'")
            return None
        except KeyError:
            log.warning(f"Error parsing RSEM counts file '{f['fn']}'")
            return None

        # Save parsed data
        if len(data) > 0:
            if f["s_name"] in self.rsem_mapped_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.rsem_mapped_data[f["s_name"]] = data
            self.add_data_source(f)
        if len(multimapping_hist) > 0:
            self.rsem_multimapping_data[f["s_name"]] = multimapping_hist

    def rsem_stats_table(self):
        """Take the parsed stats from the rsem report and add them to the
        basic stats table at the top of the report"""
        headers = {
            "alignable_percent": {
                "title": f"% Alignable, {config.read_count_prefix}",
                "description": f"% Alignable reads, {config.read_count_desc}",
                "max": 100,
                "min": 0,
                "suffix": "%",
                "scale": "YlGn",
            }
        }
        self.general_stats_addcols(self.rsem_mapped_data, headers)

    def rsem_mapped_reads_plot(self):
        """Make the rsem assignment rates plot"""

        # Plot categories
        keys = {
            "Unique": {"color": "#437bb1", "name": "Aligned uniquely to a gene"},
            "Multi": {"color": "#e63491", "name": "Aligned to multiple genes"},
            "Filtered": {"color": "#b1084c", "name": "Filtered due to too many alignments"},
            "Unalignable": {"color": "#7f0000", "name": "Unalignable reads"},
        }

        # Config for the plot
        config = {
            "id": "rsem_assignment_plot",
            "title": "RSEM: Mapped reads",
            "ylab": "# Reads",
            "cpswitch_counts_label": "Number of Reads",
            "hide_zero_cats": False,
        }

        self.add_section(
            name="Mapped Reads",
            anchor="rsem_mapped_reads",
            description="A breakdown of how all reads were aligned for each sample.",
            plot=bargraph.plot(self.rsem_mapped_data, keys, config),
        )

    def rsem_multimapping_plot(self):
        """Make a line plot showing the multimapping levels"""

        pconfig = {
            "id": "rsem_multimapping_rates",
            "title": "RSEM: Multimapping Rates",
            "ylab": "Counts",
            "xlab": "Number of alignments",
            "xDecimals": False,
            "ymin": 0,
            "tt_label": "<b>{point.x} alignments</b>: {point.y:.0f}",
        }

        self.add_section(
            name="Multimapping rates",
            anchor="rsem_multimapping",
            description="A frequency histogram showing how many reads were aligned to `n` reference regions.",
            helptext="""In an ideal world, every sequence reads would align uniquely to a single location in the
                reference. However, due to factors such as repeititve sequences, short reads and sequencing errors,
                reads can be align to the reference 0, 1 or more times. This plot shows the frequency of each factor
                of multimapping. Good samples should have the majority of reads aligning once.""",
            plot=linegraph.plot(self.rsem_multimapping_data, pconfig),
        )



""" MultiQC module to parse results from jellyfish  """


import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Jellyfish",
            anchor="jellyfish",
            href="https://github.com/gmarcais/Jellyfish",
            info="is a tool for fast, memory-efficient counting of k-mers in DNA.",
            doi="10.1093/bioinformatics/btr011",
        )

        self.jellyfish_data = dict()
        self.jellyfish_max_x = 0
        for f in self.find_log_files("jellyfish", filehandles=True):
            self.parse_jellyfish_data(f)

            # Superfluous function call to confirm that it is used in this module
            # Replace None with actual version if it is available
            self.add_software_version(None, f["s_name"])

        if self.jellyfish_max_x < 100:
            # the maximum is below 100, we display anyway up to 200
            self.jellyfish_max_x = 200
        else:
            # in this case the area plotted is a function of the maximum x
            self.jellyfish_max_x = 2 * self.max_key

        self.jellyfish_data = self.ignore_samples(self.jellyfish_data)

        if len(self.jellyfish_data) == 0:
            raise ModuleNoSamplesFound

        # Write data to file
        self.write_data_file(self.jellyfish_data, "jellyfish")

        log.info(f"Found {len(self.jellyfish_data)} reports")

        self.frequencies_plot(xmax=self.jellyfish_max_x)

    def parse_jellyfish_data(self, f):
        """Go through the hist file and memorise it"""
        histogram = {}
        occurence = 0
        for line in f["f"]:
            line = line.rstrip("\n")
            occurence = int(line.split(" ")[0])
            count = int(line.split(" ")[1])
            histogram[occurence] = occurence * count
        # delete last occurnece as it is the sum of all kmer occuring more often than it.
        del histogram[occurence]
        # sanity check
        self.max_key = max(histogram, key=histogram.get)
        self.jellyfish_max_x = max(self.jellyfish_max_x, self.max_key)
        if len(histogram) > 0:
            if f["s_name"] in self.jellyfish_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            self.add_data_source(f)
            self.jellyfish_data[f["s_name"]] = histogram

    def frequencies_plot(self, xmin=0, xmax=200):
        """Generate the qualities plot"""

        helptext = """
            A possible way to assess the complexity of a library even in
            absence of a reference sequence is to look at the kmer profile of the reads.
            The idea is to count all the kmers (_i.e._, sequence of length `k`) that occur
            in the reads. In this way it is possible to know how many kmers occur
            `1,2,.., N` times and represent this as a plot.
            This plot tell us for each x, how many k-mers (y-axis) are present in the
            dataset in exactly x-copies.

            In an ideal world (no errors in sequencing, no bias, no  repeated regions)
            this plot should be as close as  possible to a gaussian distribution.
            In reality we will always see a peak for `x=1` (_i.e._, the errors)
            and another peak close to the expected coverage. If the genome is highly
            heterozygous a second peak at half of the coverage can be expected."""

        pconfig = {
            "id": "Jellyfish_kmer_plot",
            "title": "Jellyfish: K-mer plot",
            "ylab": "Counts",
            "xlab": "k-mer frequency",
            "xDecimals": False,
            "xmin": xmin,
            "xmax": xmax,
        }

        self.add_section(
            anchor="jellyfish_kmer_plot",
            description="The K-mer plot lets you estimate library complexity and coverage from k-mer content.",
            helptext=helptext,
            plot=linegraph.plot(self.jellyfish_data, pconfig),
        )



""" MultiQC module to parse output from bustools inspect """


import json
import logging
import os

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import bargraph, table
from multiqc.utils import config

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="Bustools",
            anchor="bustools",
            href="https://bustools.github.io/",
            info="works with BUS files - a file format for single-cell RNA-seq data "
            "designed to facilitate the development of modular workflows for data "
            "processing.",
            doi="10.1093/bioinformatics/btz279",
        )

        self.prepare_data()

        log.info(f"Found {len(self.bustools_data)} logs")
        self.write_data_file(self.bustools_data, "multiqc_macs")

        self.bustools_general_stats()
        self.bustools_section()

    def prepare_data(self):
        # Parse logs
        self.bustools_data = dict()
        for f in self.find_log_files("bustools", filehandles=True):
            content = json.load(f["f"])
            s_name = self.clean_s_name(os.path.basename(f["root"]), f, root=os.path.dirname(f["root"]))
            self.bustools_data[s_name] = content
            self.add_data_source(f)

        # Filter to strip out ignored sample names
        self.bustools_data = self.ignore_samples(self.bustools_data)

        if len(self.bustools_data) == 0:
            raise ModuleNoSamplesFound

        # Superfluous function call to confirm that it is used in this module
        # Replace None with actual version if it is available
        self.add_software_version(None)

        # now fill out the table(s) headers
        self.headers = {
            "numRecords": {
                "title": f"{config.read_count_prefix} Bus Records",
                "description": f"The number of Bus Records ({config.read_count_desc})",
                "scale": "BuGn",
                "min": 0,
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "format": "{:,.2f}",
            },
            "numReads": {
                "title": f"{config.read_count_prefix} Reads",
                "description": f"Total number of reads ({config.read_count_desc})",
                "min": 0,
                "scale": "Greens",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "format": "{:,.2f}",
            },
            "numBarcodes": {
                "title": "Barcodes",
                "description": "Number of distinct barcodes",
                "min": 0,
                "scale": "YlGn",
                "format": "{:,.0f}",
                "shared_key": "barcodes",
            },
            "medianReadsPerBarcode": {
                "title": "Median reads per barcode",
                "scale": "RdYlGn",
                "min": 0,
                "format": "{:,.2f}",
            },
            "meanReadsPerBarcode": {
                "title": "Mean reads per barcode",
                "scale": "BuGn",
                "min": 0,
                "format": "{:,.2f}",
            },
            "numUMIs": {
                "title": f"{config.read_count_prefix} Distinct UMIs",
                "description": "Number of distinct Unique Molecular Identifiers (UMIs) ({})".format(
                    config.read_count_desc
                ),
                "scale": "Purples",
                "min": 0,
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "format": "{:,.2f}",
            },
            "numBarcodeUMIs": {
                "title": f"{config.read_count_prefix} Distinct barcode-UMI",
                "description": "Number of distinct barcode and Unique Molecular Identifiers (UMIs) pairs ({})".format(
                    config.read_count_desc
                ),
                "scale": "Greens",
                "min": 0,
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
                "format": "{:,.2f}",
            },
            "meanUMIsPerBarcode": {
                "title": "Mean UMIs per barcode",
                "scale": "PuBuGn",
                "min": 0,
                "format": "{:,.2f}",
            },
            "gtRecords": {
                "title": f"{config.read_count_prefix} 2xdepth records",
                "description": "Estimated number of new records at 2x sequencing depth ({})".format(
                    config.read_count_desc
                ),
                "min": 0,
                "scale": "Oranges",
                "format": "{:,.2f}",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "numTargets": {
                "title": f"{config.read_count_prefix} Distinct targets",
                "description": f"Number of distinct targets detected ({config.read_count_desc})",
                "min": 0,
                "scale": "BuGn",
                "hidden": True,
                "format": "{:,.2f}",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "meanTargetsPerSet": {
                "title": "Mean targets",
                "description": "Mean number of targets per set",
                "min": 0,
                "scale": "Greens",
                "hidden": True,
                "format": "{:,.2f}",
            },
            "numSingleton": {
                "title": f"{config.read_count_prefix} Singleton targets",
                "description": f"Number of reads with singleton target ({config.read_count_desc})",
                "min": 0,
                "scale": "Blues",
                "hidden": True,
                "format": "{:,.2f}",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "gtTargets": {
                "title": f"{config.read_count_prefix} 2xdepth targets",
                "description": "Estimated number of new targets at 2x sequencing depth ({})".format(
                    config.read_count_desc
                ),
                "min": 0,
                "scale": "BuGn",
                "hidden": True,
                "format": "{:,.2f}",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "numBarcodesOnWhitelist": {
                "title": "Whitelisted barcodes",
                "description": "Number of barcodes in agreement with whitelist",
                "min": 0,
                "scale": "Greens",
                "hidden": True,
                "format": "{:,.0f}",
                "shared_key": "barcodes",
            },
            "percentageBarcodesOnWhitelist": {
                "title": "% Whitelisted barcodes",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
            },
            "numReadsOnWhitelist": {
                "title": f"{config.read_count_prefix} Whitelisted reads",
                "description": "Number of reads with barcode in agreement with whitelist ({})".format(
                    config.read_count_desc
                ),
                "scale": "PuBu",
                "min": 0,
                "hidden": True,
                "format": "{:,.2f}",
                "shared_key": "read_count",
                "modify": lambda x: x * config.read_count_multiplier,
            },
            "percentageReadsOnWhitelist": {
                "title": "% Whitelisted reads",
                "min": 0,
                "max": 100,
                "suffix": "%",
                "scale": "RdYlGn",
            },
        }

    def bustools_general_stats(self):
        # we take a smaller subset
        general_stats_headers = [
            "numRecords",
            "numBarcodes",
            "meanReadsPerBarcode",
            "numUMIs",
            "gtRecords",
            "percentageReadsOnWhitelist",
        ]
        headers = {}
        for header, v in self.headers.items():
            if header in general_stats_headers:
                headers[header] = {key: value for key, value in v.items()}  # deep copy
                if header not in ["numRecords", "numUMIs", "percentageReadsOnWhitelist"]:
                    headers[header]["hidden"] = True
        self.general_stats_addcols(self.bustools_data, headers)

    def bustools_section(self):
        """Add bargraphs showing the mean UMIs per barcode and percentages in whitelist"""
        # add the summary table
        tconfig = {"namespace": "Bustools", "id": "bustools_summary", "table_title": "Bustools Summary Table"}
        self.add_section(
            name="Summary table",
            anchor="bustools-inspect",
            description="This is a table of the complete output of bustools inspect. Note that some columns are hidden by default (click <em>Configure Columns</em> to show).",
            plot=table.plot(self.bustools_data, self.headers, tconfig),
        )

        # also make some nice barplots
        # barplot for mean umis per sample
        mean_umis = {
            sample: {"Mean UMIs per barcode": values["meanUMIsPerBarcode"]}
            for sample, values in self.bustools_data.items()
        }

        self.add_section(
            name="Mean number of UMIs per barcode",
            anchor="bustools-umis",
            description="Average number of UMIs (unique molecular identifiers) per barcode",
            helptext="Each unique barcode represents a cell and each Unique Molecular Identifier (UMI) represents "
            "a unique transcript molecule. By counting the mean number of UMIs per barcode, you "
            "effectively calculate the average number of unique transcripts per cell.",
            plot=bargraph.plot(
                mean_umis,
                pconfig={
                    "id": "bus_umis",
                    "title": "Bustools: Mean number of UMIs per barcode per sample",
                    "cpswitch": False,
                    "tt_percentages": False,
                    "ylab": "Mean UMIs per barcode",
                },
            ),
        )

        # barplot for the percentage of reads and barcodes on the whitelist
        percentage_whitelist = {
            sample: {
                "Reads on whitelist": values["percentageReadsOnWhitelist"],
                "Barcodes on whitelist": values["percentageBarcodesOnWhitelist"],
            }
            for sample, values in self.bustools_data.items()
        }
        self.add_section(
            name="Percentage in whitelist",
            anchor="bustools-reads",
            description="The whitelist is a list of unique barcodes used in your protocol, either provided or inferred from the data.",
            helptext="Each unique barcode from the whitelist represents a cell. The percentage of "
            "reads with barcode / barcodes in the whitelist is a measure of percentage of reads that could "
            "be assigned to a cell.",
            plot=bargraph.plot(
                percentage_whitelist,
                pconfig={
                    "id": "bus_reads",
                    "title": "Bustools: Barcodes / reads with barcodes in the whitelist",
                    "ymax": 100,
                    "ymin": 0,
                    "cpswitch": False,
                    "tt_percentages": False,
                    "ylab": "Percentage of barcodes / reads with barcodes in the whitelist",
                    "stacking": None,
                    "ysuffix": "%",
                },
            ),
        )



"""MultiQC module to parse the output from deepTools"""
import logging

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound

from .bamPEFragmentSizeDistribution import bamPEFragmentSizeDistributionMixin

# deepTools modules
from .bamPEFragmentSizeTable import bamPEFragmentSizeTableMixin
from .estimateReadFiltering import EstimateReadFilteringMixin
from .plotCorrelation import plotCorrelationMixin
from .plotCoverage import plotCoverageMixin
from .plotEnrichment import PlotEnrichmentMixin
from .plotFingerprint import PlotFingerprintMixin
from .plotPCA import plotPCAMixin
from .plotProfile import plotProfileMixin

# Initialise the logger
log = logging.getLogger(__name__)


class MultiqcModule(
    BaseMultiqcModule,
    bamPEFragmentSizeTableMixin,
    bamPEFragmentSizeDistributionMixin,
    EstimateReadFilteringMixin,
    plotCoverageMixin,
    PlotEnrichmentMixin,
    PlotFingerprintMixin,
    plotProfileMixin,
    plotPCAMixin,
    plotCorrelationMixin,
):
    def __init__(self):
        # Initialise the parent object
        super(MultiqcModule, self).__init__(
            name="deepTools",
            anchor="deepTools",
            target="deepTools",
            href="http://deeptools.readthedocs.io",
            info=" is a suite of tools to process and analyze deep sequencing data.",
            doi="10.1093/nar/gkw257",
        )

        # Set up class objects to hold parsed data
        self.general_stats_headers = dict()
        self.general_stats_data = dict()
        n = dict()

        # plotCorrelation
        n["plotCorrelation"] = self.parse_plotCorrelation()
        if n["plotCorrelation"] > 0:
            log.debug(f"Found {n['plotCorrelation']} deepTools plotCorrelation samples")

        # plotPCA
        n["plotPCA"] = self.parse_plotPCA()
        if n["plotPCA"] > 0:
            log.debug(f"Found {n['plotPCA']} deepTools plotPCA samples")

        # plotEnrichment
        n["plotEnrichment"] = self.parse_plot_enrichment()
        if n["plotEnrichment"] > 0:
            log.debug(f"Found {n['plotEnrichment']} deepTools plotEnrichment samples")

        # plotFingerprint
        n["plotFingerprintOutQualityMetrics"], n["plotFingerprintOutRawCounts"] = self.parse_plotFingerprint()
        if n["plotFingerprintOutQualityMetrics"] + n["plotFingerprintOutRawCounts"] > 0:
            extra = ""
            if n["plotFingerprintOutRawCounts"] == 0:
                extra = (
                    " (you may need to increase the maximum log file size to find plotFingerprint --outRawCounts files)"
                )
            log.debug(
                "Found {} and {} deepTools plotFingerprint --outQualityMetrics and --outRawCounts samples, respectively{}".format(
                    n["plotFingerprintOutQualityMetrics"], n["plotFingerprintOutRawCounts"], extra
                )
            )

        # bamPEFragmentSizeDistribution
        n["bamPEFragmentSizeDistribution"] = self.parse_bamPEFragmentSizeDistribution()
        if n["bamPEFragmentSizeDistribution"] > 0:
            log.debug(
                "Found {} deepTools 'bamPEFragmentSize --outRawFragmentLengths' samples".format(
                    n["bamPEFragmentSizeDistribution"]
                )
            )

        # bamPEFragmentSizeTable
        n["bamPEFragmentSize"] = self.parse_bamPEFragmentSize()
        if n["bamPEFragmentSize"] > 0:
            log.debug(f"Found {n['bamPEFragmentSize']} deepTools 'bamPEFragmentSize --table' samples")

        # plotProfile
        n["plotProfile"] = self.parse_plotProfile()
        if n["plotProfile"] > 0:
            log.debug(f"Found {n['plotProfile']} deepTools plotProfile samples")

        # plotCoverage
        n["plotCoverageStdout"], n["plotCoverageOutRawCounts"] = self.parse_plotCoverage()
        if n["plotCoverageStdout"] + n["plotCoverageOutRawCounts"] > 0:
            extra = ""
            if n["plotCoverageOutRawCounts"] == 0:
                extra = (
                    " (you may need to increase the maximum log file size to find plotCoverage --outRawCounts files)"
                )
            log.debug(
                "Found {} and {} deepTools plotCoverage standard output and --outRawCounts samples, respectively{}".format(
                    n["plotCoverageStdout"], n["plotCoverageOutRawCounts"], extra
                )
            )

        # estimateReadFiltering
        n["estimateReadFiltering"] = self.parse_estimate_read_filtering()
        if n["estimateReadFiltering"] > 0:
            log.debug(f"Found {n['estimateReadFiltering']} deepTools estimateReadFiltering samples")

        tot = sum(n.values())
        if tot > 0:
            log.info(f"Found {tot} total deepTools samples")
        else:
            raise ModuleNoSamplesFound

    def _int(self, val):
        """Avoids Python3 error:
        ValueError: invalid literal for self._int() with base 10: '1.0'
        """
        return int(round(float(val)))



"""
=====
 Geo
=====

Welcome to Geo - an example template / easter egg.

Theme courtesy of the wonderful Geo for Bootstrap theme:
http://code.divshot.com/geo-bootstrap/

"""
import os

template_parent = "default"

template_dir = os.path.dirname(__file__)
base_fn = "base.html"

output_subdir = "multiqc_report"
copy_files = ["assets"]



""" MultiQC module to parse output from HiFiasm """

import logging
import re

from multiqc.modules.base_module import BaseMultiqcModule, ModuleNoSamplesFound
from multiqc.plots import linegraph

# Initialise the logger
log = logging.getLogger(__name__)

VERSION_REGEX = r"\[M::main\] Version: ([\d\.r\-]+)"


class MultiqcModule(BaseMultiqcModule):
    def __init__(self):
        # Initialse the parent object
        super(MultiqcModule, self).__init__(
            name="HiFiasm",
            anchor="hifiasm",
            href="https://github.com/chhylp123/hifiasm",
            info="is a haplotype-resolved assembler for accurate Hifi reads",
            doi="10.1038/s41592-020-01056-5",
        )

        # To store the mod data
        self.hifiasm_data = dict()
        self.parse_hifiasm_log_files()
        self.hifiasm_data = self.ignore_samples(self.hifiasm_data)

        # If we found no data
        if not self.hifiasm_data:
            raise ModuleNoSamplesFound
        log.info(f"Found {len(self.hifiasm_data)} reports")

        self.write_data_file(self.hifiasm_data, "multiqc_hifiasm_report")
        self.add_sections()

    def parse_hifiasm_log_files(self):
        for f in self.find_log_files("hifiasm", filehandles=True):
            self.add_data_source(f)
            if f["s_name"] in self.hifiasm_data:
                log.debug(f"Duplicate sample name found! Overwriting: {f['s_name']}")
            data = self.extract_kmer_graph(f["f"])
            if data:
                self.hifiasm_data[f["s_name"]] = data

            version = self.extract_version(f["f"])
            if version is not None:
                self.add_software_version(version, f["s_name"])

    def add_sections(self):
        # Plot configuration
        config = {
            "id": "hifiasm-kmer-graph",
            "title": "HiFiasm: kmer graph",
            "ylab": "Count of kmer occurrence",
            "xlab": "Kmer occurrence",
            "logswitch": True,
            "logswitch_active": True,
        }

        self.add_section(
            name="HiFiasm kmer graph",
            anchor="hifiasm-kmer-section",
            description="Kmer counts in the input data",
            helptext="""
                The kmer distribution graph for the input data. For homozygous
                samples, there should be one peak around read coverage. For
                heterozygous samples, there should be two peaks, see the
                [HiFiasm documentation](https://hifiasm.readthedocs.io/en/latest/interpreting-output.html#hifiasm-log-interpretation)
                for details.
                """,
            plot=linegraph.plot(self.hifiasm_data, config),
        )

    def extract_version(self, fin):
        """Extract the Hifiasm version from file contents"""
        for line in fin:
            if not line.startswith("[M::main]"):
                continue

            version_match = re.search(VERSION_REGEX, line)
            if version_match:
                return version_match.group(1)
        return None

    def extract_kmer_graph(self, fin):
        """Extract the kmer graph from file in"""
        data = dict()

        found_histogram = False

        for line in fin:
            if line.startswith("[M::ha_hist_line]"):
                found_histogram = True
                spline = line.strip().split()
                # Occurrence of kmer
                occurrence = spline[1][:-1]
                # Special case
                if occurrence == "rest":
                    continue
                # Count of the occurrence, checking for lines with no asterisk before count.
                if "*" in spline[2]:
                    count = int(spline[3])
                else:
                    count = int(spline[2])
                data[int(occurrence)] = count
            # If we are no longer in the histogram
            elif found_histogram:
                return data
